# -*- coding: utf-8 -*-

from AccessControl import Unauthorized
from Acquisition import aq_parent
from collective.easyform import easyformMessageFactory as _
from collective.easyform.api import get_schema
from collective.easyform.interfaces import IEasyFormFieldContext
from collective.easyform.interfaces import IEasyFormFieldsContext
from collective.easyform.interfaces import IEasyFormFieldsEditorExtender
from lxml import etree
from plone import api
from plone.schemaeditor.browser.field.edit import EditView
from plone.schemaeditor.browser.field.edit import FieldEditForm
from plone.schemaeditor.browser.field.traversal import FieldContext
from plone.schemaeditor.browser.schema.listing import SchemaListing
from plone.schemaeditor.browser.schema.listing import SchemaListingPage
from plone.schemaeditor.browser.schema.traversal import SchemaContext
from plone.supermodel.parser import SupermodelParseError
from Products.CMFPlone.utils import safe_bytes
from Products.Five import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from z3c.form import button
from zope.cachedescriptors.property import Lazy as lazy_property
from zope.component import getAdapters
from zope.component import queryMultiAdapter
from zope.interface import implementer
from ZPublisher.BaseRequest import DefaultPublishTraverse
from Products.statusmessages.interfaces import IStatusMessage
from Products.CMFPlone.utils import safe_unicode

import html

NAMESPACE = "{http://namespaces.plone.org/supermodel/schema}"

try:
    from plone.schemaeditor import SchemaEditorMessageFactory as __
except ImportError:
    from plone.schemaeditor import _ as __

try:
    import plone.resourceeditor

    plone.resourceeditor  # avoid PEP 8 warning
    HAVE_RESOURCE_EDITOR = True
except ImportError:  # pragma: no cover
    HAVE_RESOURCE_EDITOR = False


@implementer(IEasyFormFieldContext)
class EasyFormFieldContext(FieldContext):
    """Wrapper for published zope 3 schema fields."""


@implementer(IEasyFormFieldsContext)
class EasyFormFieldsView(SchemaContext):

    schema = None

    def __init__(self, context, request):
        self.schema = get_schema(context)
        super(EasyFormFieldsView, self).__init__(self.schema, request, name="fields")

    def publishTraverse(self, request, name):
        """Look up the field whose name matches the next URL path element,
        and wrap it.
        """
        try:
            return EasyFormFieldContext(self.schema[name], self.request).__of__(self)
        except KeyError:
            return DefaultPublishTraverse(self, request).publishTraverse(request, name)

    def browserDefault(self, request):
        """If not traversing through the schema to a field, show the
        SchemaListingPage.
        """
        return self, ("@@listing",)

    @property
    def allowedFields(self):
        fields = api.portal.get_registry_record("easyform.allowedFields")
        return fields


class FieldsSchemaListing(SchemaListing):
    template = ViewPageTemplateFile("fields_listing.pt")

    @property
    def default_fieldset_label(self):
        return (
            self.context.aq_parent.default_fieldset_label
            or super(FieldsSchemaListing, self).default_fieldset_label
        )

    def handleModelEdit(self, action):
        self.request.response.redirect("@@modeleditor")

    @button.buttonAndHandler(
        _(u'Save'),
    )
    def handleSave(self, action):
        super(FieldsSchemaListing, self).handleSaveDefaults(self, action)
        return self.request.RESPONSE.redirect(aq_parent(self.context).absolute_url())


if HAVE_RESOURCE_EDITOR:
    but = button.Button("modeleditor", title=_(u"Edit XML Fields Model"))
    FieldsSchemaListing.buttons += button.Buttons(but)
    handler = button.Handler(but, FieldsSchemaListing.handleModelEdit)
    FieldsSchemaListing.handlers.addHandler(but, handler)


class EasyFormFieldsListingPage(SchemaListingPage):
    """Form wrapper so we can get a form with layout.

    We define an explicit subclass rather than using the wrap_form method
    from plone.z3cform.layout so that we can inject the schema name into
    the form label.
    """

    form = FieldsSchemaListing
    index = ViewPageTemplateFile("model_listing.pt")


class FieldEditForm(FieldEditForm):
    @lazy_property
    def additionalSchemata(self):
        schema_context = self.context.aq_parent
        adapters = getAdapters(
            (schema_context, self.field), IEasyFormFieldsEditorExtender
        )
        return [v for k, v in adapters]


class EditView(EditView):
    form = FieldEditForm


class ModelEditorView(BrowserView):
    """Editor view.
    Mostly stolen from plone.app.dexterity.browser.modeleditor.ModelEditorView
    """

    template = ViewPageTemplateFile("modeleditor.pt")

    title = _(u"Edit XML Fields Model")

    def modelSource(self):
        return self.context.aq_parent.fields_model

    def authorized(self, context, request):
        authenticator = queryMultiAdapter((context, request), name=u"authenticator")
        return authenticator and authenticator.verify()

    def save(self, source):
        self.context.aq_parent.fields_model = source

    def __call__(self):
        """View and eventually save the form."""

        save = "form.button.save" in self.request.form
        source = self.request.form.get("source")
        if save and source:

            # First, check for authenticator
            if not self.authorized(self.context, self.request):
                raise Unauthorized

            # Is it valid XML?
            # Some safety measures.
            # We do not want to load entities, especially file:/// entities.
            # Also discard processing instructions.
            #
            source = safe_bytes(source)
            parser = etree.XMLParser(resolve_entities=False, remove_pis=True)
            try:
                root = etree.fromstring(source, parser=parser)
            except etree.XMLSyntaxError as e:
                IStatusMessage(self.request).addStatusMessage(
                    "XMLSyntaxError: {0}".format(html.escape(safe_unicode(e.args[0]))),
                    "error",
                )
                return super().__call__()

            # a little more sanity checking, look at first two element levels
            if root.tag != NAMESPACE + "model":
                IStatusMessage(self.request).addStatusMessage(
                    _(u"Error: root tag must be 'model'"),
                    "error",
                )
                return super().__call__()

            for element in root.getchildren():
                if element.tag != NAMESPACE + "schema":
                    IStatusMessage(self.request).addStatusMessage(
                        _(u"Error: all model elements must be 'schema'"),
                        "error",
                    )
                    return super().__call__()

            # can supermodel parse it?
            # This is mainly good for catching bad dotted names.
            try:
                plone.supermodel.loadString(source, policy=u"dexterity")
            except SupermodelParseError as e:
                message = e.args[0].replace('\n  File "<unknown>"', "")
                IStatusMessage(self.request).addStatusMessage(
                    u"SuperModelParseError: {0}".format(html.escape(message)),
                    "error",
                )
                return super().__call__()

            # clean up formatting sins
            source = etree.tostring(
                root, pretty_print=True, xml_declaration=True, encoding="utf8"
            )
            # and save
            self.save(source)
        # import pdb; pdb.set_trace()
        return self.template()



# -*- coding: utf-8 -*-
#
# Integeration tests specific to the mailer
#
from collective.easyform.api import get_actions
from collective.easyform.api import get_context
from collective.easyform.api import get_schema
from collective.easyform.api import set_actions
from collective.easyform.api import set_fields
from collective.easyform.interfaces import IActionExtender
from collective.easyform.tests import base
from email.header import decode_header
from importlib import import_module
from plone import api
from plone.app.textfield.value import RichTextValue
from plone.namedfile.file import NamedFile
from Products.CMFPlone.utils import safe_unicode
from six import BytesIO

import datetime
import unittest


try:
    # Python 3
    from email import message_from_bytes

    LINESEP = b"\r\n"
except ImportError:
    # Python 2
    from email import message_from_string as message_from_bytes

    LINESEP = b"\n"

try:
    from openpyxl import load_workbook
    HAS_OPENPYXL = True
except ImportError:
    HAS_OPENPYXL = False


HAS_PLONE_6 = getattr(
    import_module("Products.CMFPlone.factory"), "PLONE60MARKER", False
)


MODEL_WITH_ALL_FIELDS = """
<model xmlns:i18n="http://xml.zope.org/namespaces/i18n" xmlns:marshal="http://namespaces.plone.org/supermodel/marshal" xmlns:form="http://namespaces.plone.org/supermodel/form" xmlns:security="http://namespaces.plone.org/supermodel/security" xmlns:users="http://namespaces.plone.org/supermodel/users" xmlns:lingua="http://namespaces.plone.org/supermodel/lingua" xmlns:easyform="http://namespaces.plone.org/supermodel/easyform" xmlns="http://namespaces.plone.org/supermodel/schema">
  <schema>
    <field name="replyto" type="zope.schema.TextLine" easyform:TDefault="python:member and member.getProperty('email', '') or ''" easyform:serverSide="False" easyform:validators="isValidEmail">
      <description/>
      <title>Your E-Mail Address</title>
    </field>
    <field name="topic" type="zope.schema.TextLine">
      <description/>
      <title>Subject</title>
    </field>
    <field name="comments" type="zope.schema.Text" easyform:serverSide="False" easyform:THidden="False">
      <description/>
      <title>Comments</title>
    </field>
    <field name="richtext" type="plone.app.textfield.RichText">
      <description/>
      <required>False</required>
      <title>richtext</title>
    </field>
    <field name="datetime" type="zope.schema.Datetime">
      <description/>
      <required>False</required>
      <title>datetime</title>
    </field>
    <field name="date" type="zope.schema.Date">
      <description/>
      <required>False</required>
      <title>date</title>
    </field>
    <field name="bool" type="zope.schema.Bool">
      <description/>
      <required>False</required>
      <title>bool</title>
      <form:widget type="z3c.form.browser.radio.RadioFieldWidget"/>
    </field>
    <field name="number" type="zope.schema.Int">
      <description/>
      <required>False</required>
      <title>number</title>
    </field>
    <field name="floating" type="zope.schema.Float">
      <description/>
      <required>False</required>
      <title>floating</title>
    </field>
    <field name="tuple" type="zope.schema.Tuple">
      <description/>
      <required>False</required>
      <title>tuple</title>
      <value_type type="zope.schema.Choice">
        <values/>
      </value_type>
    </field>
    <field name="list" type="zope.schema.List">
      <description/>
      <required>False</required>
      <title>list</title>
      <value_type type="zope.schema.Choice">
        <values/>
      </value_type>
    </field>
    <field name="empty_string" type="zope.schema.TextLine">
      <description/>
      <required>False</required>
      <title>empty_string</title>
    </field>
  </schema>
</model>
"""

class TestFunctions(base.EasyFormTestCase):
    """Test mailer action"""

    def dummy_send(self, mfrom, mto, messageText, immediate=False):
        self.mfrom = mfrom
        self.mto = mto
        if hasattr(messageText, "encode"):
            # It is text instead of bytes.
            messageText = messageText.encode("utf-8")
        self.messageText = messageText
        TWOLINESEP = LINESEP + LINESEP
        self.messageBody = TWOLINESEP.join(messageText.split(TWOLINESEP)[1:])

    def afterSetUp(self):
        super(TestFunctions, self).afterSetUp()
        self.folder.invokeFactory("EasyForm", "ff1")
        self.ff1 = getattr(self.folder, "ff1")
        self.ff1.CSRFProtection = False  # no csrf protection
        self.mailhost = self.folder.MailHost
        self.mailhost._send = self.dummy_send
        actions = get_actions(self.ff1)
        actions["mailer"].recipient_email = u"mdummy@address.com"
        set_actions(self.ff1, actions)

    def LoadRequestForm(self, **kwargs):
        request = self.layer["request"]
        request.form.clear()
        prefix = "form.widgets."
        for key in kwargs.keys():
            request.form[prefix + key] = kwargs[key]
        return request

    def test_DummyMailer(self):
        """sanity check; make sure dummy mailer works as expected"""

        self.mailhost.send(
            "messageText", mto="dummy@address.com", mfrom="dummy1@address.com"
        )
        self.assertTrue(self.messageText.endswith(b"messageText"))
        self.assertEqual(self.mto, ["dummy@address.com"])
        self.assertIn(b"To: dummy@address.com", self.messageText)
        self.assertEqual(self.mfrom, "dummy1@address.com")
        self.assertIn(b"From: dummy1@address.com", self.messageText)

    def test_Mailer_Basic(self):
        """Test mailer with dummy_send"""

        mailer = get_actions(self.ff1)["mailer"]

        data = {"topic": "test subject", "replyto": "foo@spam.com", "comments": "test comments"}
        request = self.LoadRequestForm(**data)

        mailer.onSuccess(data, request)

        self.assertIn(b"To: mdummy@address.com", self.messageText)
        self.assertIn(b"Subject: =?utf-8?q?test_subject?=", self.messageText)
        msg = message_from_bytes(self.messageText)
        self.assertIn("test comments", msg.get_payload(decode=False))

    def test_MailerAdditionalHeaders(self):
        """Test mailer with dummy_send"""

        mailer = get_actions(self.ff1)["mailer"]

        data = {"topic": "test subject", "replyto": "foo@spam.com", "comments": "test comments"}
        request = self.LoadRequestForm(**data)

        mailer.additional_headers = ["Generator: Plone", "Token:   abc  "]

        mailer.onSuccess(data, request)

        self.assertIn(b"Generator: Plone", self.messageText)
        self.assertIn(b"Token: abc", self.messageText)
        self.assertIn(b"To: mdummy@address.com", self.messageText)
        self.assertIn(b"Subject: =?utf-8?q?test_subject?=", self.messageText)
        msg = message_from_bytes(self.messageText)
        self.assertIn("test comments", msg.get_payload(decode=False))

    def test_MailerLongSubject(self):
        """Test mailer with subject line > 76 chars (Tracker # 84)"""

        long_subject = (
            "Now is the time for all good persons to come to"
            "the aid of the quick brown fox."
        )

        mailer = get_actions(self.ff1)["mailer"]

        data = {"topic": long_subject, "replyto": "foo@spam.com", "comments": "test comments"}
        request = self.LoadRequestForm(**data)
        mailer.onSuccess(data, request)

        msg = message_from_bytes(self.messageText)
        encoded_subject_header = msg["subject"]
        decoded_header = decode_header(encoded_subject_header)[0][0]

        self.assertEqual(decoded_header, long_subject.encode("utf-8"))

    def test_SubjectDollarReplacement(self):
        """
        Simple subject lines should do ${identifier} replacement from
        request.form -- but only for a basic override.
        """
        mailer = get_actions(self.ff1)["mailer"]
        mailer.msg_subject = "This is my ${topic} now"
        data = dict(
            topic="test subject", replyto="test@test.org", comments="test comments"
        )

        # baseline unchanged
        request = self.LoadRequestForm(**data)
        self.messageText = ""
        mailer.onSuccess(data, request)
        self.assertIn(b"Subject: =?utf-8?q?test_subject?=", self.messageText)

        data2 = dict(
            topic="test ${subject}", replyto="test@test.org", comments="test comments"
        )

        # no substitution on field replacement (default situation)
        request = self.LoadRequestForm(**data2)
        self.messageText = ""
        mailer.onSuccess(data2, request)
        self.assertIn(b"Subject: =?utf-8?q?test_=24=7Bsubject=7D?=", self.messageText)

        # we should get substitution in a basic override
        mailer.subject_field = ""
        request = self.LoadRequestForm(**data)
        self.messageText = ""
        mailer.onSuccess(data, request)
        self.assertIn(
            b"Subject: =?utf-8?q?This_is_my_test_subject_now?=", self.messageText
        )

        # we should get substitution in a basic override
        mailer.msg_subject = "This is my ${untopic} now"
        self.messageText = b""
        mailer.onSuccess(data, request)
        self.assertIn(
            b"Subject: =?utf-8?q?This_is_my_=3F=3F=3F_now?=", self.messageText
        )

        # we don't want substitution on user input
        request = self.LoadRequestForm(**data2)
        self.messageText = b""
        mailer.onSuccess(data2, request)
        self.assertIn(
            b"Subject: =?utf-8?q?This_is_my_=3F=3F=3F_now?=", self.messageText
        )

    def test_TemplateReplacement(self):
        """
        Mail template prologues, epilogues and footers should do ${identifier}
        replacement from request.form -- this is simpler because there are no
        overrides.
        """

        mailer = get_actions(self.ff1)["mailer"]
        mailer.body_pre = "Hello ${topic},"
        mailer.body_post = "Thanks, ${topic}!"
        mailer.body_footer = "Eat my footer, ${topic}."
        data = dict(
            topic="test subject", replyto="test@test.org", comments="test comments"
        )

        # we should get substitution
        request = self.LoadRequestForm(**data)
        self.messageText = b""
        mailer.onSuccess(data, request)

        self.assertIn(b"Hello test subject,", self.messageBody)
        self.assertIn(b"Thanks, test subject!", self.messageBody)
        self.assertIn(b"Eat my footer, test subject.", self.messageBody)

    def test_UTF8Subject(self):
        """Test mailer with uft-8 encoded subject line"""

        utf8_subject = u"Effacer les entrées sauvegardées"
        data = dict(
            topic=utf8_subject, replyto="test@test.org", comments="test comments"
        )

        mailer = get_actions(self.ff1)["mailer"]
        # fields = self.ff1._getFieldObjects()
        request = self.LoadRequestForm(**data)
        mailer.onSuccess(data, request)

        msg = message_from_bytes(self.messageText)
        encoded_subject_header = msg["subject"]
        decoded_header = decode_header(encoded_subject_header)[0][0]

        self.assertEqual(safe_unicode(decoded_header), utf8_subject)

    def test_UnicodeSubject(self):
        """Test mailer with Unicode encoded subject line"""
        utf8_subject = u"Effacer les entrées sauvegardées"
        unicode_subject = utf8_subject
        data = dict(
            topic=unicode_subject, replyto="test@test.org", comments="test comments"
        )

        mailer = get_actions(self.ff1)["mailer"]
        request = self.LoadRequestForm(**data)
        mailer.onSuccess(data, request)

        msg = message_from_bytes(self.messageText)
        encoded_subject_header = msg["subject"]
        decoded_header = decode_header(encoded_subject_header)[0][0]

        self.assertEqual(safe_unicode(decoded_header), utf8_subject)

    def test_Utf8ListSubject(self):
        """Test mailer with Unicode encoded subject line"""
        utf8_subject_list = [u"Effacer les entrées", u"sauvegardées"]
        data = dict(
            topic=utf8_subject_list, replyto="test@test.org", comments="test comments"
        )

        mailer = get_actions(self.ff1)["mailer"]
        request = self.LoadRequestForm(**data)
        mailer.onSuccess(data, request)

        msg = message_from_bytes(self.messageText)
        encoded_subject_header = msg["subject"]
        decoded_header = decode_header(encoded_subject_header)[0][0]

        self.assertEqual(safe_unicode(decoded_header), ", ".join(utf8_subject_list))

    def test_MailerOverrides(self):
        """Test mailer override functions"""

        mailer = get_actions(self.ff1)["mailer"]
        mailer.subjectOverride = "python: '{0} and {1}'.format('eggs', 'spam')"
        mailer.senderOverride = "string: spam@eggs.com"
        mailer.recipientOverride = "string: eggs@spam.com"
        data = {"topic": "test subject", "replyto": "foo@spam.com", "comments": "test comments"}
        request = self.LoadRequestForm(**data)

        mailer.onSuccess(data, request)
        self.assertIn(b"Subject: =?utf-8?q?eggs_and_spam?=", self.messageText)
        self.assertIn(b"From: spam@eggs.com", self.messageText)
        self.assertIn(b"To: eggs@spam.com", self.messageText)

    def test_MailerOverridesWithFieldValues(self):
        mailer = get_actions(self.ff1)["mailer"]
        mailer.subjectOverride = "fields/topic"
        mailer.recipientOverride = "fields/replyto"
        data = {"topic": "eggs and spam", "replyto": "test@test.ts", "comments": "test comments"}
        request = self.LoadRequestForm(**data)
        mailer.onSuccess(data, request)

        self.assertIn(b"Subject: =?utf-8?q?eggs_and_spam?=", self.messageText)
        self.assertIn(b"To: test@test.ts", self.messageText)

    def testMultiRecipientOverrideByString(self):
        """try multiple recipients in recipient override"""

        mailer = get_actions(self.ff1)["mailer"]
        mailer.recipientOverride = "string: eggs@spam.com, spam@spam.com"

        data = {"topic": "test subject", "replyto": "foo@spam.com", "comments": "cool stuff"}
        request = self.LoadRequestForm(**data)
        mailer.onSuccess(data, request)

        self.assertIn(b"To: eggs@spam.com, spam@spam.com", self.messageText)

    def testMultiRecipientOverrideByTuple(self):
        """try multiple recipients in recipient override"""

        mailer = get_actions(self.ff1)["mailer"]
        mailer.recipientOverride = "python: ('eggs@spam.com', 'spam.spam.com')"

        data = {"topic": "test subject", "replyto": "foo@spam.com", "comments": "cool stuff"}
        request = self.LoadRequestForm(**data)
        mailer.onSuccess(data, request)

        self.assertIn(b"To: eggs@spam.com, spam.spam.com", self.messageText)

    def testRecipientFromRequest(self):
        """try recipient from designated field"""

        mailer = get_actions(self.ff1)["mailer"]
        mailer.to_field = "replyto"
        mailer.replyto_field = None

        fields = {"topic": "test subject", "replyto": "eggs@spamandeggs.com", "comments": "cool stuff"}

        request = self.LoadRequestForm(**fields)
        mailer.onSuccess(fields, request)

        self.assertIn(b"To: eggs@spamandeggs.com", self.messageText)

        fields = {
            "topic": "test subject",
            "replyto": ["eggs@spam.com", "spam@spam.com"],
            "comments": "cool stuff",
        }
        request = self.LoadRequestForm(**fields)
        mailer.onSuccess(fields, request)

        self.assertTrue(self.messageText.find(b"To: eggs@spam.com, spam@spam.com") > 0)

    def setExecCondition(self, value):
        actions = get_actions(self.ff1)
        IActionExtender(actions["mailer"]).execCondition = value
        set_actions(self.ff1, actions)

    def test_ExecConditions(self):
        """Test mailer with various exec conditions"""

        # if an action adapter's execCondition is filled in and evaluates
        # false, the action adapter should not fire.

        view = self.ff1.restrictedTraverse("view")
        form = view.form_instance

        fields = dict(
            topic="test subject", replyto="test@test.org", comments="test comments"
        )
        self.LoadRequestForm(**fields)

        self.messageText = b""
        self.setExecCondition("python: False")
        form.processActions(fields)
        self.assertTrue(len(self.messageText) == 0)

        self.messageText = b""
        self.setExecCondition("python: True")
        form.processActions(fields)
        self.assertTrue(len(self.messageText) > 0)

        self.messageText = b""
        self.setExecCondition("python: 1==0")
        form.processActions(fields)
        self.assertTrue(len(self.messageText) == 0)

        # make sure an empty execCondition causes the action to fire
        self.messageText = b""
        self.setExecCondition("")
        form.processActions(fields)
        self.assertTrue(len(self.messageText) > 0)

    def test_selectiveFieldMailing(self):
        """Test selective inclusion of fields in the mailing"""

        mailer = get_actions(self.ff1)["mailer"]
        fields = dict(
            topic="test subject", replyto="test@test.org", comments="test comments"
        )
        request = self.LoadRequestForm(**fields)

        # make sure all fields are sent unless otherwise specified
        self.messageText = b""
        mailer.onSuccess(fields, request)
        self.assertIn(b"te=" + LINESEP + b"st subject", self.messageBody)
        self.assertIn(b"test@test.org", self.messageBody)
        self.assertIn(b"test comments", self.messageBody)

        # setting some show fields shouldn't change that
        mailer.showFields = ("topic", "comments")
        self.messageText = b""
        mailer.onSuccess(fields, request)
        self.assertIn(b"te=" + LINESEP + b"st subject", self.messageBody)
        self.assertIn(b"test@test.org", self.messageBody)
        self.assertIn(b"test comments", self.messageBody)

        # until we turn off the showAll flag
        mailer.showAll = False
        self.messageText = b""
        mailer.onSuccess(fields, request)
        self.assertIn(b"te=" + LINESEP + b"st subject", self.messageBody)
        self.assertNotIn(b"test@test.org", self.messageBody)
        self.assertIn(b"test comments", self.messageBody)

        # check includeEmpties
        mailer.includeEmpties = False

        # first see if everything's still included
        mailer.showAll = True
        self.messageText = b""
        mailer.onSuccess(fields, request)
        # look for labels
        self.assertTrue(
            self.messageBody.find(b"Subject") > 0
            and self.messageBody.find(b"Your E-Mail Address") > 0
            and self.messageBody.find(b"Comments") > 0
        )

        # now, turn off required for a field and leave it empty
        fields = get_schema(self.ff1)
        fields["comments"].required = False
        set_fields(self.ff1, fields)
        fields = {"topic": "test subject", "replyto": "test@test.org", "comments": ""}
        request = self.LoadRequestForm(**fields)
        self.messageText = b""
        mailer.onSuccess(fields, request)
        self.assertIn(b"Subject", self.messageBody)
        self.assertIn(b"Your E-Mail Address", self.messageBody)
        self.assertNotIn(b"Comments", self.messageBody)

    def test_ccOverride(self):
        """Test override for CC field"""

        mailer = get_actions(self.ff1)["mailer"]
        fields = dict(
            topic="test subject", replyto="test@test.org", comments="test comments"
        )
        request = self.LoadRequestForm(**fields)
        mailer.cc_recipients = "test@testme.com"
        self.messageText = b""
        mailer.onSuccess(fields, request)
        self.assertIn("test@testme.com", self.mto)

        # simple override
        mailer.ccOverride = "string:test@testme.com"
        self.messageText = b""
        mailer.onSuccess(fields, request)
        self.assertIn("test@testme.com", self.mto)

        # list override
        mailer.ccOverride = "python:['test@testme.com', 'test1@testme.com']"
        self.messageText = b""
        mailer.onSuccess(fields, request)
        self.assertTrue(
            "test@testme.com" in self.mto and "test1@testme.com" in self.mto
        )

    def test_bccOverride(self):
        """Test override for BCC field"""
        mailer = get_actions(self.ff1)["mailer"]
        fields = dict(
            topic="test subject", replyto="test@test.org", comments="test comments"
        )
        request = self.LoadRequestForm(**fields)
        mailer.bcc_recipients = "test@testme.com"
        self.messageText = b""
        mailer.onSuccess(fields, request)
        self.assertIn("test@testme.com", self.mto)

        # simple override
        mailer.bccOverride = "string:test@testme.com"
        self.messageText = b""
        mailer.onSuccess(fields, request)
        self.assertIn("test@testme.com", self.mto)

        # list override
        mailer.bccOverride = "python:['test@testme.com', 'test1@testme.com']"
        self.messageText = b""
        mailer.onSuccess(fields, request)
        self.assertTrue(
            "test@testme.com" in self.mto and "test1@testme.com" in self.mto
        )

    def testNoRecipient(self):
        """
        This is not the feature in easyform as we need recipient in the mailer
        for easyforms. It will not take the default site's recipient.
        """
        mailer = get_actions(self.ff1)["mailer"]
        mailer.recipient_email = u""
        mailer.to_field = None
        mailer.replyto_field = None
        fields = dict(
            topic="test subject", replyto="test@test.org", comments="test comments"
        )

        request = self.LoadRequestForm(**fields)

        self.messageText = b""
        self.assertRaises(ValueError, mailer.onSuccess, fields, request)

    def test_custom_email_template(self):
        """Test mailer with custom template"""
        default_fields = api.content.create(
            self.portal, "File", id="easyform_mail_body_default.pt"
        )
        default_fields.file = NamedFile("Custom e-mail template!")
        fields = dict(
            topic="test subject", replyto="test@test.org", comments="test comments"
        )
        request = self.LoadRequestForm(**fields)

        mailer = get_actions(self.ff1)["mailer"]
        mailer.onSuccess(fields, request)
        self.assertIn(b"Custom e-mail template!", self.messageText)

    def test_MailerXMLAttachments(self):
        """Test mailer with dummy_send"""
        mailer = get_actions(self.ff1)["mailer"]
        mailer.sendXML = True
        mailer.sendCSV = False
        context = get_context(mailer)
        context.fields_model = MODEL_WITH_ALL_FIELDS
        # Test all dexterity field type listed at https://docs.plone.org/external/plone.app.dexterity/docs/reference/fields.html
        fields = dict(
            replyto="test@test.org",
            topic="test subject",
            richtext="Raw",
            comments=u"test comments😀",
            datetime="2019-04-01T00:00:00",
            date="2019-04-02",
            delta=datetime.timedelta(1),
            bool=True,
            number=1981,
            floating=3.14,
            tuple=("elemenet1", "element2"),
            list=[1, 2, 3, 4],
            empty_string="",
        )
        if not HAS_PLONE_6:
            fields["datetime"] = "2019-04-01 00:00:00"
        request = self.LoadRequestForm(**fields)
        attachments = mailer.get_attachments(fields, request)
        self.assertEqual(1, len(attachments))
        self.assertIn(
            u"Content-Type: application/xml\nMIME-Version: 1.0\nContent-Transfer-Encoding: base64\nContent-Disposition: attachment",
            mailer.get_mail_text(fields, request, context),
        )
        name, mime, enc, xml = attachments[0]
        output_nodes = [
            b'<field name="replyto">test@test.org</field>',
            b'<field name="topic">test subject</field>',
            b'<field name="richtext">Raw</field>',
            b'<field name="comments">test comments\xf0\x9f\x98\x80</field>',
            b'<field name="datetime">2019-04-01T00:00:00</field>',
            b'<field name="date">2019-04-02</field>',
            b'<field name="bool">True</field>',
            b'<field name="number">1981</field>',
            b'<field name="floating">3.14</field>',
            b'<field name="tuple">["elemenet1", "element2"]</field>',
            b'<field name="list">["1", "2", "3", "4"]</field>',
            b'<field name="empty_string" />',
        ]
        if not HAS_PLONE_6:
            output_nodes[4] = b'<field name="datetime">2019-04-01 00:00:00</field>'

        self.assertIn(b"<?xml version='1.0' encoding='utf-8'?>\n<form>", xml)

        # the order of the nodes can change ... check each line
        for node in output_nodes:
            self.assertIn(node, xml)

    def test_MailerCSVAttachments(self):
        """Test mailer with dummy_send"""
        mailer = get_actions(self.ff1)["mailer"]
        mailer.sendXML = False
        mailer.sendCSV = True
        context = get_context(mailer)
        context.fields_model = MODEL_WITH_ALL_FIELDS
        # Test all dexterity field type listed at https://docs.plone.org/external/plone.app.dexterity/docs/reference/fields.html
        fields = dict(
            replyto="test@test.org",
            topic="test subject",
            richtext="Raw",
            comments=u"test comments😀",
            datetime="2019-04-01T00:00:00",
            date="2019-04-02",
            delta=datetime.timedelta(1),
            bool=True,
            number=1981,
            floating=3.14,
            tuple=("elemenet1", "element2"),
            list=[1, 2, 3, 4],
            empty_string="",
        )
        if not HAS_PLONE_6:
            fields["datetime"] = "2019-04-01 00:00:00"
        request = self.LoadRequestForm(**fields)

        attachments = mailer.get_attachments(fields, request)
        self.assertEqual(1, len(attachments))
        self.assertIn(
            u"Content-Type: application/csv\nMIME-Version: 1.0\nContent-Transfer-Encoding: base64\nContent-Disposition: attachment",
            mailer.get_mail_text(fields, request, context),
        )
        name, mime, enc, csv = attachments[0]
        output = [
            b"test@test.org",
            b"test subject",
            b"Raw",
            b"test comments\xf0\x9f\x98\x80",
            b"2019-04-01T00:00:00",
            b"2019-04-02",
            b"True",
            b"1981",
            b"3.14",
            b'[""elemenet1"", ""element2""]',
            b'[""1"", ""2"", ""3"", ""4""]',
            b"",
        ]
        if not HAS_PLONE_6:
            output[4] = b"2019-04-01 00:00:00"

        # the order of the columns can change ... check each
        # TODO should really have a header row
        for value in output:
            self.assertIn(value, csv)

    @unittest.skipUnless(HAS_OPENPYXL, "Requires openpyxl")
    def test_MailerXLSXAttachments(self):
        """ Test mailer with dummy_send """
        mailer = get_actions(self.ff1)["mailer"]
        mailer.sendXML = False
        mailer.sendCSV = False
        mailer.sendXLSX = True
        context = get_context(mailer)
        context.fields_model = MODEL_WITH_ALL_FIELDS
        # Test all dexterity field type listed at https://docs.plone.org/external/plone.app.dexterity/docs/reference/fields.html
        fields = dict(
            replyto="test@test.org",
            topic="test subject",
            richtext="Raw",
            comments=u"test comments😀",
            datetime="2019-04-01T00:00:00",
            date="2019-04-02",
            delta=datetime.timedelta(1),
            bool=True,
            number=1981,
            floating=3.14,
            tuple=("elemenet1", "element2"),
            list=[1, 2, 3, 4],
            empty_string="",
        )
        if not HAS_PLONE_6:
            fields["datetime"] = "2019-04-01 00:00:00"

        request = self.LoadRequestForm(**fields)
        attachments = mailer.get_attachments(fields, request)
        self.assertEqual(1, len(attachments))
        self.assertIn(
            u"Content-Type: application/vnd.openxmlformats-officedocument.spreadsheetml.sheet\nMIME-Version: 1.0\nContent-Transfer-Encoding: base64\nContent-Disposition: attachment",
            mailer.get_mail_text(fields, request, context),
        )
        name, mime, enc, xlsx = attachments[0]
        wb = load_workbook(BytesIO(xlsx))
        wb.active
        ws = wb.active

        row = [cell.value and cell.value.encode('utf-8') or b'' for cell in list(ws.rows)[0]]

        output = [
            b"test@test.org",
            b"test subject",
            b"Raw",
            b"test comments\xf0\x9f\x98\x80",
            b"2019-04-01T00:00:00",
            b"2019-04-02",
            b"True",
            b"1981",
            b"3.14",
            b'["elemenet1", "element2"]',
            b'["1", "2", "3", "4"]',
            b"",
        ]
        if not HAS_PLONE_6:
            output[4] = b"2019-04-01 00:00:00"

        # the order of the columns can change ... check each
        # TODO should really have a header row
        for value in output:
            self.assertIn(value, row)



# -*- coding: utf-8 -*-
from AccessControl import getSecurityManager
from Acquisition import aq_base
from Acquisition import aq_chain
from Acquisition import aq_parent
from collections import OrderedDict
from collective.easyform import easyformMessageFactory as _
from collective.easyform.api import dollar_replacer
from collective.easyform.api import filter_fields
from collective.easyform.api import get_actions
from collective.easyform.api import get_expression
from collective.easyform.api import get_schema
from collective.easyform.config import FORM_ERROR_MARKER
from collective.easyform.interfaces import IActionExtender
from collective.easyform.interfaces import IEasyFormForm
from collective.easyform.interfaces import IEasyFormThanksPage
from collective.easyform.interfaces import IEasyFormWidget
from collective.easyform.interfaces import IFieldExtender
from collective.easyform.interfaces import ISaveData
from logging import getLogger
from os.path import splitext
from plone.app.z3cform.inline_validation import InlineValidationView
from plone.autoform.form import AutoExtensibleForm
from plone.namedfile.interfaces import INamed
from plone.z3cform.layout import FormWrapper
from Products.Five import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from z3c.form import button
from z3c.form import form
from z3c.form.interfaces import DISPLAY_MODE
from z3c.form.interfaces import HIDDEN_MODE
from z3c.form.interfaces import IErrorViewSnippet
from zope.component import getMultiAdapter
from zope.i18nmessageid import MessageFactory
from zope.interface import alsoProvides
from zope.interface import implementer
from zope.schema import getFieldsInOrder
from zope.schema import ValidationError
from ZPublisher.mapply import mapply

import six


logger = getLogger("collective.easyform")
PMF = MessageFactory("plone")

try:
    from Products.CMFPlone.utils import safe_encode
except ImportError:
    # only thing needed to maintain 5.0.x compatibility
    def safe_bytes(value, encoding="utf-8"):
        """Convert text to bytes of the specified encoding."""
        if isinstance(value, six.text_type):
            value = value.encode(encoding)
        return value

    safe_encode = safe_bytes


@implementer(IEasyFormForm)
class EasyFormForm(AutoExtensibleForm, form.Form):
    """EasyForm form."""

    form_template = ViewPageTemplateFile("easyform_form.pt")
    thank_you_template = ViewPageTemplateFile("thank_you.pt")
    ignoreContext = True
    css_class = "easyformForm row"
    thanksPage = False

    # allow prefill - see Products/CMFPlone/patches/z3c_form
    allow_prefill_from_GET_request = True

    @property
    def method(self):
        return self.context.method

    @property
    def prologue(self):
        return self.context.prologue

    @property
    def epilogue(self):
        return self.context.epilogue

    @property
    def default_fieldset_label(self):
        return (
            self.context.default_fieldset_label
            or super(EasyFormForm, self).default_fieldset_label
        )

    def action(self):
        """Redefine <form action=''> attribute."""
        action = getattr(self.context, "formActionOverride")
        if action:
            action = get_expression(self.context, action)
        if not action:
            action = self.context.absolute_url()
        if self.context.forceSSL:
            action = action.replace("http://", "https://")
        return action

    @property
    def enable_form_tabbing(self):
        return self.context.form_tabbing

    @property
    def enable_unload_protection(self):
        return self.context.unload_protection

    @property
    def enableCSRFProtection(self):
        return self.context.CSRFProtection

    @property
    def label(self):
        return self.thanksPage and self.context.thankstitle or self.context.Title()

    @property
    def description(self):
        return (
            self.thanksPage
            and self.context.thanksdescription
            or self.context.Description()
        )

    @property
    def schema(self):
        return get_schema(self.context)

    def updateServerSideData(self, data):
        for fname in self.schema:
            field = self.schema[fname]
            efield = IFieldExtender(field)
            serverSide = getattr(efield, "serverSide", False)
            if not serverSide:
                continue
            fdefault = field.default
            TDefault = getattr(efield, "TDefault", None)
            value = get_expression(self.context, TDefault) if TDefault else fdefault
            data[fname] = value
        return data

    def processActions(self, fields):
        # get a list of adapters with no duplicates, retaining order
        actions = getFieldsInOrder(get_actions(self.context))
        for name, action in actions:
            if not action.required:
                continue
            # Now, see if we should execute it.
            # Check to see if execCondition exists and has contents
            execCondition = IActionExtender(action).execCondition
            if execCondition:
                doit = get_expression(self.context, execCondition)
            else:
                doit = True
            if doit and hasattr(action, "onSuccess"):
                result = action.onSuccess(fields, self.request)
                if isinstance(result, dict) and len(result):
                    return result

    def setErrorsMessage(self, errors):
        for field in errors:
            if field not in self.widgets:
                if field == FORM_ERROR_MARKER:
                    self.formErrorsMessage = errors[field]
                continue
            error = ValidationError()
            error.doc = lambda: errors[field]
            view = getMultiAdapter(
                (
                    error,
                    self.request,
                    self.widgets[field],
                    self.widgets[field].field,
                    self,
                    self.context,
                ),
                IErrorViewSnippet,
            )
            view.update()
            self.widgets.errors += (view,)
            self.widgets[field].error = view
        self.status = self.formErrorsMessage

    @button.buttonAndHandler(
        PMF(u"Submit"), name="submit", condition=lambda form: not form.thanksPage
    )
    def handleSubmit(self, action):
        unsorted_data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        unsorted_data = self.updateServerSideData(unsorted_data)
        errors = self.processActions(unsorted_data)
        if errors:
            return self.setErrorsMessage(errors)
        data = OrderedDict(
            [x for x in getFieldsInOrder(self.schema) if x[0] in unsorted_data]
        )
        data.update(unsorted_data)
        thanksPageOverride = self.context.thanksPageOverride
        if not thanksPageOverride:
            # we come back to the form itself.
            # the thanks page is handled in the __call__ method
            return
        thanksPageOverrideAction = self.context.thanksPageOverrideAction
        thanksPage = get_expression(self.context, thanksPageOverride)
        if six.PY2 and isinstance(thanksPage, six.text_type):
            thanksPage = thanksPage.encode("utf-8")
        if thanksPageOverrideAction == "redirect_to":
            self.request.response.redirect(thanksPage)
            return
        if thanksPageOverrideAction == "traverse_to":
            thanksPage = self.context.restrictedTraverse(thanksPage)
            thanksPage = mapply(thanksPage, self.request.args, self.request)
            self.request.response.write(safe_encode(thanksPage))

    @button.buttonAndHandler(
        _(u"Reset"), name="reset", condition=lambda form: form.context.useCancelButton
    )
    def handleReset(self, action):
        self.request.response.redirect(self.nextURL())

    def nextURL(self):
        return self.context.absolute_url()

    def setOmitFields(self, fields):
        omit = []
        new_fields = []
        for fname, field in fields.items():
            efield = IFieldExtender(field.field)
            TEnabled = getattr(efield, "TEnabled", None)
            serverSide = getattr(efield, "serverSide", False)
            if TEnabled and not get_expression(self.context, TEnabled) or serverSide:
                omit.append(fname)
            if getattr(efield, "THidden", False):
                field.mode = HIDDEN_MODE
                field.field.ignoreRequest = False
            new_fields.append(field)
        fields = fields.__class__(*new_fields)
        if omit:
            fields = fields.omit(*omit)
        return fields

    def setThanksFields(self, fields, data):
        omit = filter_fields(self.context, self.schema, data, omit=True)
        new_fields = []
        for fname, field in fields.items():
            if field.mode == HIDDEN_MODE:
                field.mode = DISPLAY_MODE
            new_fields.append(field)
        fields = fields.__class__(*new_fields)
        if omit:
            fields = fields.omit(*omit)

        return fields

    def set_depends_on(self, fields):
        for fname, field in fields.items():
            efield = IFieldExtender(field.field)
            depends_on = getattr(efield, "depends_on", None)
            if depends_on:
                field.field.setTaggedValue("depends_on", depends_on)
        return fields

    def set_css_class(self, fields):
        for fname, field in fields.items():
            efield = IFieldExtender(field.field)
            css_class = getattr(efield, "css_class", None)
            if css_class:
                field.field.setTaggedValue("css_class", css_class)
        return fields

    def updateFields(self):
        if self.thanksPage:
            return
        super(EasyFormForm, self).updateFields()
        if not hasattr(self, "base_fields"):
            self.base_fields = self.fields
        if not hasattr(self, "base_groups"):
            self.base_groups = dict([(i.label, i.fields) for i in self.groups])
        self.fields = self.setOmitFields(self.base_fields)
        self.fields = self.set_depends_on(self.fields)
        self.fields = self.set_css_class(self.fields)
        for group in self.groups:
            group.fields = self.setOmitFields(self.base_groups.get(group.label))
            group.fields = self.set_depends_on(group.fields)
            group.fields = self.set_css_class(group.fields)

    def updateActions(self):
        super(EasyFormForm, self).updateActions()
        if "submit" in self.actions:
            if self.context.submitLabelOverride:
                self.actions["submit"].title = get_expression(
                    self.context, self.context.submitLabelOverride
                )
            else:
                self.actions["submit"].title = self.context.submitLabel
        if "reset" in self.actions:
            self.actions["reset"].title = self.context.resetLabel

    def markWidgets(self):
        for w in self.widgets.values():
            if not IEasyFormWidget.providedBy(w):
                alsoProvides(w, IEasyFormWidget)
        for g in self.groups:
            for w in g.widgets.values():
                if not IEasyFormWidget.providedBy(w):
                    alsoProvides(w, IEasyFormWidget)

    def formMaybeForceSSL(self):
        """Redirect to an https:// URL if the 'force SSL' option is on.

        However, don't do so for those with rights to edit the form,
        to avoid making the form uneditable if SSL isn't configured
        properly.  These users will still get an SSL-ified form
        action for when the form is submitted.
        """
        sm = getSecurityManager()
        if self.context.forceSSL and not sm.checkPermission(
            "cmf.ModifyPortalContent", self.context
        ):
            # Make sure we're being accessed via a secure connection
            if self.request["SERVER_URL"].startswith("http://"):
                secure_url = self.request["URL"].replace("http://", "https://")
                self.request.response.redirect(secure_url, status="movedtemporarily")

    def update(self):
        """Update form - see interfaces.IForm"""
        self.formMaybeForceSSL()
        super(EasyFormForm, self).update()
        self.markWidgets()
        self.template = self.form_template
        if self.request.method != "POST" or self.context.thanksPageOverride:
            # go with all but default thank you page rendering
            return
        # If an adapter has already set errors, don't re-run extraction and
        # validation, just bail out:
        # (we copy the logic from plone.app.z3cform at templates/macros.pt)
        if self.widgets.errors or self.status == getattr(
            self, "formErrorsMessage", None
        ):
            return
        data, errors = self.extractData()
        if errors:
            # render errors
            return
        data = self.updateServerSideData(data)
        self.thanksPage = True
        self.template = self.thank_you_template
        self.fields = self.setThanksFields(self.base_fields, data)
        for name in list(self.widgets.keys()):
            if name not in self.fields:
                del self.widgets[name]
        self.widgets.mode = self.mode = DISPLAY_MODE
        self.widgets.update()

        for group in self.groups:
            group.fields = self.setThanksFields(self.base_groups.get(group.label), data)
            for name in list(group.widgets.keys()):
                if name not in group.fields:
                    del group.widgets[name]
            group.widgets.mode = DISPLAY_MODE
            group.widgets.update()
        prologue = self.context.thanksPrologue
        epilogue = self.context.thanksEpilogue
        self.thanksPrologue = prologue and dollar_replacer(prologue.output, data)
        self.thanksEpilogue = epilogue and dollar_replacer(epilogue.output, data)
        alsoProvides(self.request, IEasyFormThanksPage)

    def header_injection(self):
        tal_expression = self.context.headerInjection
        if not tal_expression:
            return ""
        header_to_inject = get_expression(self.context, tal_expression)
        if six.PY2 and isinstance(header_to_inject, six.text_type):
            header_to_inject = header_to_inject.encode("utf-8")

        return header_to_inject

    def form_name(self):
        if self.context.nameAttribute:
            return self.context.nameAttribute
        return None


class EasyFormFormWrapper(FormWrapper):
    form = EasyFormForm
    index = ViewPageTemplateFile("easyform_layout.pt")

    def header_injection(self):
        header_injection = self.form_instance.header_injection()
        return header_injection

    def css_class(self):
        css_class = None
        if self.form_instance.thanksPage:
            css_class = u"easyform-thankspage"
        return css_class


EasyFormView = EasyFormFormWrapper


class EasyFormFormEmbedded(EasyFormForm):
    """EasyForm form embedded."""

    form_template = ViewPageTemplateFile("easyform_form_embedded.pt")


class EasyFormInlineValidationView(InlineValidationView):
    def __call__(self, fname=None, fset=None):
        self.context = EasyFormForm(self.context, self.request)
        return super(EasyFormInlineValidationView, self).__call__(fname, fset)


class GetSaveDataAdaptersView(BrowserView):
    def __call__(self, *args, **kwargs):
        """Return all contained save data adapters."""
        view = EasyFormForm(self.context, self.request)
        form = view.context
        adapters = []
        actions = get_actions(form)
        for action_id in actions:
            action = actions[action_id]
            if ISaveData.providedBy(action):
                adapters.append(action)
        return adapters


class ValidateFile(BrowserView):
    def __call__(self, value, size=1048576, allowed_types=None, forbidden_types=None):
        if not value:
            return False
        if not INamed.providedBy(value):
            return False
        if size and value.getSize() > size:
            return _(
                "msg_file_too_big",
                mapping={"size": size},
                default=u"File is bigger than allowed size of ${size} bytes!",
            )
        ftype = splitext(value.filename)[-1]
        # remove leading dot '.' from file extension
        ftype = ftype and ftype[1:].lower() or ""
        if allowed_types and ftype not in allowed_types:
            return _(
                "msg_file_not_allowed",
                mapping={"ftype": ftype.upper()},
                default=u'File type "${ftype}" is not allowed!',
            )
        if forbidden_types and ftype in forbidden_types:
            return _(
                "msg_file_not_allowed",
                mapping={"ftype": ftype.upper()},
                default=u'File type "${ftype}" is not allowed!',
            )
        return False


class GetEasyFormURL(BrowserView):
    """Helper view for calculating the right url in actions.xml.

    A url like /fields or /actions needs to be called an the easyform object.
    Danger is that you get a url like easyform/fields/fields.
    This is because /fields and /actions may look like browser views,
    but to Plone they look like content items.
    This messes up some logic in plone_context_state.

    Usage in a url_expr:

      python:context.restrictedTraverse('@@get-easyform-url')('fields')

    That will give the correct url:

      easyform-object-url/fields

    """

    def __call__(self, name=""):
        # First get the real easyform, if we can find it.
        form = self.get_form()
        if form is None:
            # We did not find the form.  View is called on the wrong item.
            # Give back a relative link anyway.
            base = "."
            if not name.startswith("/"):
                name = "/" + name
            return base + name
        base = form.absolute_url()
        if not name:
            # Do not add a needless slash at the end.
            return base
        if not name.startswith("/"):
            name = "/" + name
        return base + name

    def get_form(self):
        for item in aq_chain(self.context):
            # Note that the fields/actions contexts also have
            # portal_type EasyForm, but this is due to acquisition.
            base = aq_base(item)
            if not hasattr(base, "portal_type"):
                continue
            if base.portal_type == "EasyForm":
                return item
            if base.portal_type == "Plone Site":
                # We have gone too far already.
                return


class IsSubEasyForm(GetEasyFormURL):
    """Is this a sub object of an easyform?

    For use in actions.xml.

    If this is a sub object of easyform,
    then if has portal_type EasyForm,
    but only due to acquisition.
    """

    def __call__(self):
        if self.context.portal_type != "EasyForm":
            # The wrong portal_type.
            return False
        if hasattr(aq_base(self.context), "portal_type"):
            # An actual EasyForm
            return False
        return True


class FolderContentsView(BrowserView):
    """folder_contents view for EasyForm.

    When you are at url easyform/fields the Contents link in the toolbar
    points to easyform/folder_contents, which does not exist.
    So let's redirect to the parent.
    """

    def __call__(self):
        url = aq_parent(self.context).absolute_url() + "/folder_contents"
        self.request.response.redirect(url)
        return url


# BBB
ValidateFileSize = ValidateFile



from plone import api

import logging


logger = logging.getLogger(__name__)


def update_last_compilation(context, timetuple=None):
    # Let's do the imports inline, so they are not needlessly done at startup.
    # Should not really matter, but oh well.
    from datetime import datetime
    from plone.registry.interfaces import IRegistry
    from Products.CMFPlone.interfaces import IBundleRegistry
    from zope.component import getUtility

    registry = getUtility(IRegistry)
    records = registry.forInterface(IBundleRegistry, prefix="plone.bundles/easyform")
    # Technically we only need year, month and day.
    # But keep this in sync with registry.xml.
    records.last_compilation = datetime(*timetuple)
    logger.info("Updated the last_compilation date of the easyform bundle.")

    # Run the combine-bundles import step or its handler.
    # Can be done with an upgrade step:
    #   <gs:upgradeDepends
    #     title="combine bundles"
    #     import_steps="combine-bundles"
    #     run_deps="false"
    #     />
    # But directly calling the basic function works fine.
    # See also comment here:
    # https://github.com/collective/collective.easyform/pull/248#issuecomment-689365240
    # Also, here we can do a try/except so it does not fail on Plone 5.0,
    # which I think does not have the import step, not the function.
    try:
        from Products.CMFPlone.resources.browser.combine import combine_bundles
    except ImportError:
        logger.warning("Could not call combine_bundles. You should do that yourself.")
        return
    portal = api.portal.get()
    combine_bundles(portal)


def update_last_compilation_1007(context):
    update_last_compilation(context, (2020, 9, 8, 14, 2, 0))


def update_last_compilation_1008(context):
    update_last_compilation(context, (2020, 12, 9, 14, 2, 0))


def update_last_compilation_1009(context):
    update_last_compilation(context, (2021, 8, 31, 0, 0, 0))


def fix_savedata_persistence_issues(context):

    from persistent.mapping import PersistentMapping

    catalog = api.portal.get_tool("portal_catalog")
    forms = catalog.unrestrictedSearchResults(portal_type='EasyForm')
    for item in forms:
        form = item.getObject()
        if hasattr(form, '_inputStorage'):
            # Convert to persistent mapping
            form._inputStorage = PersistentMapping(form._inputStorage)
            logger.info(
                'Fixed storage of {}'.format('/'.join(form.getPhysicalPath()))
            )


def change_saveddata_action_permission(context):
    portal_actions = api.portal.get_tool("portal_actions")
    category = portal_actions.get("object_buttons")
    if category is None:
        return
    action = category.get("saveddata")
    if action is None:
        return
    action._setPropValue('permissions', ("collective.easyform: Download Saved Input",))


def remove_migrate_all_forms_record(context):
    from plone.registry.interfaces import IRegistry
    from zope.component import getUtility

    registry = getUtility(IRegistry)
    record = "easyform.migrate_all_forms"
    if record in registry.records:
        del registry.records[record]
        logger.info("Removed {} registry record".format(record))



# -*- coding: utf-8 -*-
from collective.easyform.tests import base
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.registry.interfaces import IRegistry
from zope.component import getUtility


try:
    from plone.testing.zope import Browser
except ImportError:
    from plone.testing.z2 import Browser


class ControlPanelTestCase(base.EasyFormTestCase):
    """Test that changes in the easyform control panel are actually
    stored in the registry.
    """

    def setUp(self):
        self.portal_url = self.layer["portal"].absolute_url()
        self.browser = Browser(self.layer["app"])
        self.browser.handleErrors = False
        self.browser.addHeader(
            "Authorization", "Basic " + SITE_OWNER_NAME + ":" + SITE_OWNER_PASSWORD
        )

    def test_easyform_control_panel_link_to_overview(self):
        self.browser.open(self.portal_url + "/@@overview-controlpanel")
        link = self.browser.getLink("easyform")
        self.assertEqual(link.url, "http://nohost/plone/@@easyform-controlpanel")

    def test_easyform_control_panel_contents(self):
        self.browser.open(self.portal_url + "/@@easyform-controlpanel")
        self.assertTrue("easyform Settings" in self.browser.contents)
        self.assertTrue("Allowed Fields" in self.browser.contents)
        self.assertTrue("CSV delimiter" in self.browser.contents)
        input = self.browser.getControl(label="CSV delimiter")
        self.assertEqual(input.value, ",")

    def test_easyform_control_panel_sidebar(self):
        self.browser.open(self.portal_url + "/@@easyform-controlpanel")
        self.assertTrue("General" in self.browser.contents)
        link = self.browser.getLink("Add-ons")
        self.assertEqual(link.url, "http://nohost/plone/prefs_install_products_form")


class ControlPanelFunctionalTestCase(base.EasyFormFunctionalTestCase):
    """Test that changes in the easyform control panel are actually
    stored in the registry.
    """

    def setUp(self):
        self.portal_url = self.layer["portal"].absolute_url()
        self.browser = Browser(self.layer["app"])
        self.browser.handleErrors = False
        self.browser.addHeader(
            "Authorization", "Basic " + SITE_OWNER_NAME + ":" + SITE_OWNER_PASSWORD
        )

    def test_easyform_control_panel_allowed_fields_saved(self):
        self.browser.open(self.portal_url + "/@@easyform-controlpanel")
        self.browser.getControl("Rich Text").selected = False
        self.browser.getControl("Save").click()
        registry = getUtility(IRegistry)
        self.assertNotIn(
            "plone.app.textfield.RichText",
            registry.records["easyform.allowedFields"].value,
        )

    def test_easyform_control_panel_CSV_delimiter_saved(self):
        registry = getUtility(IRegistry)
        self.assertEqual(
            registry.records["easyform.csv_delimiter"].value,
            ",",
        )
        self.browser.open(self.portal_url + "/@@easyform-controlpanel")
        input = self.browser.getControl(label="CSV delimiter")
        input.value = ";"
        self.browser.getControl("Save").click()
        self.assertEqual(
            registry.records["easyform.csv_delimiter"].value,
            ";",
        )

    def test_easyform_control_panel_CSV_delimiter_limited_to_one_char(self):
        registry = getUtility(IRegistry)
        self.assertEqual(
            registry.records["easyform.csv_delimiter"].value,
            ",",
        )
        self.browser.open(self.portal_url + "/@@easyform-controlpanel")
        input = self.browser.getControl(label="CSV delimiter")
        input.value = "comma"
        self.browser.getControl("Save").click()
        self.assertTrue("Value is too long" in self.browser.contents)
        self.assertEqual(
            registry.records["easyform.csv_delimiter"].value,
            ",",
        )

    def test_easyform_control_panel_CSV_delimiter_required(self):
        registry = getUtility(IRegistry)
        self.assertEqual(
            registry.records["easyform.csv_delimiter"].value,
            ",",
        )
        self.browser.open(self.portal_url + "/@@easyform-controlpanel")
        input = self.browser.getControl(label="CSV delimiter")
        input.value = ""
        self.browser.getControl("Save").click()
        self.assertTrue("Required input is missing." in self.browser.contents)
        input = self.browser.getControl(label="CSV delimiter")
        self.assertEqual(input.value, "")
        self.assertEqual(
            registry.records["easyform.csv_delimiter"].value,
            ",",
        )


# EOF



# -*- coding: utf-8 -*-
from collective.easyform.interfaces import IAction
from collective.easyform.interfaces import IActionExtender
from collective.easyform.interfaces import IEasyFormActionsContext
from collective.easyform.interfaces import IEasyFormFieldsContext
from collective.easyform.interfaces import IFieldExtender
from plone.autoform.interfaces import WIDGETS_KEY
from plone.supermodel.parser import IFieldMetadataHandler
from plone.supermodel.utils import ns
from zope.component import adapter
from zope.interface import implementer
from zope.schema.interfaces import IField


@adapter(IEasyFormFieldsContext, IField)
def get_field_extender(context, field):
    return IFieldExtender


def _get_(self, key):
    return self.field.interface.queryTaggedValue(key, {}).get(self.field.__name__)


def _set_(self, value, key):
    data = self.field.interface.queryTaggedValue(key, {})
    data[self.field.__name__] = value
    self.field.interface.setTaggedValue(key, data)


@implementer(IFieldExtender)
@adapter(IField)
class FieldExtender(object):
    def __init__(self, field):
        self.field = field

    field_widget = property(
        lambda x: _get_(x, WIDGETS_KEY), lambda x, value: _set_(x, value, WIDGETS_KEY)
    )
    TDefault = property(
        lambda x: _get_(x, "TDefault"), lambda x, value: _set_(x, value, "TDefault")
    )
    TEnabled = property(
        lambda x: _get_(x, "TEnabled"), lambda x, value: _set_(x, value, "TEnabled")
    )
    depends_on = property(
        lambda x: _get_(x, "depends_on"), lambda x, value: _set_(x, value, "depends_on")
    )
    css_class = property(
        lambda x: _get_(x, "css_class"), lambda x, value: _set_(x, value, "css_class")
    )
    TValidator = property(
        lambda x: _get_(x, "TValidator"), lambda x, value: _set_(x, value, "TValidator")
    )
    serverSide = property(
        lambda x: _get_(x, "serverSide"), lambda x, value: _set_(x, value, "serverSide")
    )
    validators = property(
        lambda x: _get_(x, "validators"), lambda x, value: _set_(x, value, "validators")
    )
    THidden = property(
        lambda x: _get_(x, "THidden"), lambda x, value: _set_(x, value, "THidden")
    )


@implementer(IFieldMetadataHandler)
class EasyFormFieldMetadataHandler(object):
    """Support the easyform: namespace in model definitions."""

    namespace = "http://namespaces.plone.org/supermodel/easyform"
    prefix = "easyform"

    def read(self, fieldNode, schema, field):
        name = field.__name__
        for i in ["TDefault", "TEnabled", "TValidator", "depends_on", "css_class",]:
            value = fieldNode.get(ns(i, self.namespace))
            if value:
                data = schema.queryTaggedValue(i, {})
                data[name] = value
                schema.setTaggedValue(i, data)
        # serverSide
        value = fieldNode.get(ns("serverSide", self.namespace))
        if value:
            data = schema.queryTaggedValue("serverSide", {})
            data[name] = value.lower() == "true"
            schema.setTaggedValue("serverSide", data)
        # validators
        value = fieldNode.get(ns("validators", self.namespace))
        if value:
            data = schema.queryTaggedValue("validators", {})
            data[name] = value.split("|")
            schema.setTaggedValue("validators", data)
        # hidden
        value = fieldNode.get(ns("THidden", self.namespace))
        if value:
            data = schema.queryTaggedValue("THidden", {})
            data[name] = value.lower() == "true"
            schema.setTaggedValue("THidden", data)

    def write(self, fieldNode, schema, field):
        name = field.__name__
        for i in ["TDefault", "TEnabled", "TValidator", "depends_on", "css_class",]:
            value = schema.queryTaggedValue(i, {}).get(name, None)
            if value:
                fieldNode.set(ns(i, self.namespace), value)
        # serverSide
        value = schema.queryTaggedValue("serverSide", {}).get(name, None)
        if isinstance(value, bool):
            fieldNode.set(ns("serverSide", self.namespace), str(value))
        # validators
        value = schema.queryTaggedValue("validators", {}).get(name, None)
        if value:
            fieldNode.set(ns("validators", self.namespace), "|".join(value))
        # hidden
        value = schema.queryTaggedValue("THidden", {}).get(name, None)
        if isinstance(value, bool):
            fieldNode.set(ns("THidden", self.namespace), str(value))


@adapter(IEasyFormActionsContext, IAction)
def get_action_extender(context, action):
    return IActionExtender


@implementer(IActionExtender)
@adapter(IAction)
class ActionExtender(object):
    def __init__(self, field):
        self.field = field

    execCondition = property(
        lambda x: _get_(x, "execCondition"),
        lambda x, value: _set_(x, value, "execCondition"),
    )


@implementer(IFieldMetadataHandler)
class EasyFormActionMetadataHandler(object):
    """Support the easyform: namespace in model definitions."""

    namespace = "http://namespaces.plone.org/supermodel/easyform"
    prefix = "easyform"

    def read(self, fieldNode, schema, field):
        name = field.__name__
        value = fieldNode.get(ns("execCondition", self.namespace))
        data = schema.queryTaggedValue("execCondition", {})
        if value:
            data[name] = value
            schema.setTaggedValue("execCondition", data)

    def write(self, fieldNode, schema, field):
        name = field.__name__
        value = schema.queryTaggedValue("execCondition", {}).get(name, None)
        if value:
            fieldNode.set(ns("execCondition", self.namespace), value)



# -*- coding: utf-8 -*-
from .actions import IAction
from collective.easyform import easyformMessageFactory as _
from plone.autoform import directives
from plone.z3cform.interfaces import IFormWrapper
from z3c.form.browser.checkbox import CheckBoxFieldWidget
from zope.interface import Interface

import zope.interface
import zope.schema.interfaces


class ISavedDataFormWrapper(IFormWrapper):
    pass


class IExtraData(Interface):
    dt = zope.schema.TextLine(
        title=_(u"Posting Date/Time"), required=False, default=u"", missing_value=u""
    )
    HTTP_X_FORWARDED_FOR = zope.schema.TextLine(
        title=_(
            u"extra_header",
            default=u"${name} Header",
            mapping={u"name": u"HTTP_X_FORWARDED_FOR"},
        ),
        required=False,
        default=u"",
        missing_value=u"",
    )
    REMOTE_ADDR = zope.schema.TextLine(
        title=_(
            u"extra_header",
            default=u"${name} Header",
            mapping={u"name": u"REMOTE_ADDR"},
        ),
        required=False,
        default=u"",
        missing_value=u"",
    )
    HTTP_USER_AGENT = zope.schema.TextLine(
        title=_(
            u"extra_header",
            default=u"${name} Header",
            mapping={u"name": u"HTTP_USER_AGENT"},
        ),
        required=False,
        default=u"",
        missing_value=u"",
    )


class ISaveData(IAction):

    """A form action adapter that will save form input data and
    return it in csv- or tab-delimited format."""

    showFields = zope.schema.List(
        title=_(u"label_savefields_text", default=u"Saved Fields"),
        description=_(
            u"help_savefields_text",
            default=u"Pick the fields whose inputs you'd like to include in "
            u"the saved data. If empty, all fields will be saved.",
        ),
        unique=True,
        required=False,
        value_type=zope.schema.Choice(vocabulary="easyform.Fields"),
    )
    directives.widget(ExtraData=CheckBoxFieldWidget)
    ExtraData = zope.schema.List(
        title=_(u"label_savedataextra_text", default="Extra Data"),
        description=_(
            u"help_savedataextra_text",
            default=u"Pick any extra data you'd like saved with the form " u"input.",
        ),
        unique=True,
        value_type=zope.schema.Choice(vocabulary="easyform.ExtraDataDL"),
    )
    DownloadFormat = zope.schema.Choice(
        title=_(u"label_downloadformat_text", default=u"Download Format"),
        default=u"csv",
        vocabulary="easyform.FormatDL",
    )
    UseColumnNames = zope.schema.Bool(
        title=_(u"label_usecolumnnames_text", default=u"Include Column Names"),
        description=_(
            u"help_usecolumnnames_text",
            default=u"Do you wish to have column names on the first line of "
            u"downloaded input?",
        ),
        default=True,
        required=False,
    )



# -*- coding: utf-8 -*-
from collections import namedtuple
from collective.easyform.interfaces.mailer import default_mail_body
from collective.easyform.migration.fields import append_field
from collective.easyform.migration.fields import append_list_node
from collective.easyform.migration.fields import append_node
from collective.easyform.migration.fields import set_attribute
from collective.easyform.migration.fields import to_text
from lxml import etree
from Products.PloneFormGen.content.actionAdapter import FormActionAdapter

import logging


logger = logging.getLogger("collective.easyform.migration")


NAMESPACES = {
    "easyform": "http://namespaces.plone.org/supermodel/easyform",
    "form": "http://namespaces.plone.org/supermodel/form",
}


def append_body_pt(field, name, value):
    node = append_node(field, name, value)
    # Replace mailer page template with the default one from easyform as it's
    # unlikely that the PFG template would work.
    node.text = default_mail_body()


Type = namedtuple("Type", ["name", "handler"])
Property = namedtuple("Property", ["name", "handler"])


TYPES_MAPPING = {
    "FormMailerAdapter": Type("collective.easyform.actions.Mailer", append_field),
    "FormSaveDataAdapter": Type("collective.easyform.actions.SaveData", append_field),
    "FormCustomScriptAdapter": Type(
        "collective.easyform.actions.CustomScript", append_field
    ),
}

PROPERTIES_MAPPING = {
    "additional_headers": Property("additional_headers", append_node),
    "bcc_recipients": Property("bcc_recipients", append_node),
    "bccOverride": Property("bccOverride", append_node),
    "body_footer": Property("body_footer", append_node),
    "body_post": Property("body_post", append_node),
    "body_pre": Property("body_pre", append_node),
    "body_pt": Property("body_pt", append_body_pt),
    "body_type": Property("body_type", append_node),
    "cc_recipients": Property("cc_recipients", append_node),
    "ccOverride": Property("ccOverride", append_node),
    "description": Property("description", append_node),
    "execCondition": Property("easyform:execCondition", set_attribute),
    "msg_subject": Property("msg_subject", append_node),
    "recipient_email": Property("recipient_email", append_node),
    "recipient_name": Property("recipient_name", append_node),
    "recipientOverride": Property("recipientOverride", append_node),
    "replyto_field": Property("replyto_field", append_node),
    "senderOverride": Property("senderOverride", append_node),
    "showAll": Property("showAll", append_node),
    "showFields": Property("showFields", append_list_node),
    "subject_field": Property("subject_field", append_node),
    "subjectOverride": Property("subjectOverride", append_node),
    "title": Property("title", append_node),
    "to_field": Property("to_field", append_node),
    "xinfo_headers": Property("xinfo_headers", append_node),
    "ExtraData": Property("ExtraData", append_list_node),
    "DownloadFormat": Property("DownloadFormat", append_node),
    "UseColumnNames": Property("UseColumnNames", append_node),
    "SavedFormInput": Property("SavedFormInput", append_node),
    "ProxyRole": Property("ProxyRole", append_node),
    "ScriptBody": Property("ScriptBody", append_node),
}


def pfg_actions(context):
    actions = []
    for obj in context.objectValues():
        if not isinstance(obj, FormActionAdapter):
            continue
        id_ = obj.getId()
        props = {}
        props["_portal_type"] = obj.portal_type
        for field in obj.Schema().fields():
            name = field.getName()
            accessor = field.getEditAccessor(obj)
            props[name] = accessor()
        actions.append((id_, props))
    return actions


def actions_model(ploneformgen):
    """Create an actions xml model from a PloneFormGen instance."""
    parser = etree.XMLParser(remove_blank_text=True)
    model = etree.fromstring(ACTIONS_MODEL, parser)
    schema = model.find("{http://namespaces.plone.org/supermodel/schema}schema")

    for actionname, properties in pfg_actions(ploneformgen):
        type_ = TYPES_MAPPING.get(properties["_portal_type"])
        if type_ is None:
            # ploneformgen has been renamed with _MIGRATION_ suffix
            # hence, we remove 11 chars
            pfgpath = "/".join(ploneformgen.getPhysicalPath())[:-11]
            logger.warning(
                "Ignoring field '%s' of type '%s' of form '%s'.",
                actionname,
                properties["_portal_type"],
                pfgpath,
            )
            continue

        field = type_.handler(schema, type_.name, actionname, properties)

        for name, value in properties.items():
            prop = PROPERTIES_MAPPING.get(name)
            if prop is None:
                continue

            value = to_text(value)

            prop.handler(field, prop.name, value)

    return etree.tostring(model, pretty_print=True)


ACTIONS_MODEL = b"""
<model xmlns="http://namespaces.plone.org/supermodel/schema">
  <schema></schema>
</model>
"""



# -*- coding: utf-8 -*-

from collective.easyform.interfaces import ILabelWidget
from collective.easyform.interfaces import IRenderWidget
from collective.easyform.interfaces import IRichLabelWidget
from Products.Five.browser import BrowserView
from Products.Five.browser.metaconfigure import ViewMixinForTemplates
from z3c.form import interfaces
from z3c.form.browser import widget
from z3c.form.interfaces import IWidget
from z3c.form.widget import FieldWidget
from z3c.form.widget import Widget
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.component import adapter
from zope.interface import implementer
from zope.interface import implementer_only
from zope.interface import Interface
from zope.publisher.interfaces.browser import IBrowserView
from zope.schema.interfaces import IField


@implementer_only(ILabelWidget)
class LabelWidget(widget.HTMLFormElement, Widget):
    """Textarea widget implementation."""

    klass = u"label-widget"
    css = u"label"
    value = u""

    def update(self):
        super(LabelWidget, self).update()
        widget.addFieldClass(self)


@adapter(IField, interfaces.IFormLayer)
@implementer(interfaces.IFieldWidget)
def LabelFieldWidget(field, request):
    """IFieldWidget factory for TextWidget."""
    return FieldWidget(field, LabelWidget(request))


@implementer_only(IRichLabelWidget)
class RichLabelWidget(widget.HTMLFormElement, Widget):
    """Textarea widget implementation."""

    klass = u"rich-label-widget"
    css = u"richlabel"
    value = u""

    def update(self):
        super(RichLabelWidget, self).update()
        widget.addFieldClass(self)


@adapter(IField, interfaces.IFormLayer)
@implementer(interfaces.IFieldWidget)
def RichLabelFieldWidget(field, request):
    """IFieldWidget factory for TextWidget."""
    return FieldWidget(field, RichLabelWidget(request))


class LabelRenderWidget(ViewMixinForTemplates, BrowserView):
    index = ViewPageTemplateFile("label.pt")


class RichLabelRenderWidget(ViewMixinForTemplates, BrowserView):
    index = ViewPageTemplateFile("rich_label.pt")


# overriding plone.app.z3cform widget.pt:
@implementer(IRenderWidget)
class RenderWidget(ViewMixinForTemplates, BrowserView):
    index = ViewPageTemplateFile('widget.pt')


@adapter(IRenderWidget, Interface)
@implementer(IBrowserView)
class WidgetDependencyView(object):
    def __init__(self, widget, request):
        self.widget = widget
        self.request = request

    def __call__(self):
        field = self.widget.context.field
        if not field:
            return ""
        depends_on = field.queryTaggedValue("depends_on")
        if not depends_on:
            return ""
        return depends_on

@adapter(IRenderWidget, Interface)
@implementer(IBrowserView)
class WidgetCssClassView(object):
    def __init__(self, widget, request):
        self.widget = widget
        self.request = request

    def __call__(self):
        field = self.widget.context.field
        # default bottom margin (see plone.app.z3cform)
        dflt = "mb-3"
        if not field:
            return dflt
        css_class = field.queryTaggedValue("css_class")
        if not css_class:
            return dflt
        return css_class



# -*- coding: utf-8 -*-
from __future__ import print_function

from plone import api
from plone.app.contenttypes.testing import PLONE_APP_CONTENTTYPES_FIXTURE
from plone.app.robotframework.testing import REMOTE_LIBRARY_BUNDLE_FIXTURE
from plone.app.testing import FunctionalTesting
from plone.app.testing import IntegrationTesting
from plone.app.testing import PloneSandboxLayer
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from Products.MailHost.interfaces import IMailHost
from Products.MailHost.MailHost import MailHost
from unittest import TestCase
from zope.component import getSiteManager


try:
    from plone.testing.zope import WSGI_SERVER_FIXTURE
except ImportError:
    from plone.testing.z2 import ZSERVER_FIXTURE as WSGI_SERVER_FIXTURE

try:
    # Python 3
    from email import message_from_bytes
except ImportError:
    # Python 2
    from email import message_from_string as message_from_bytes


class MailHostMock(MailHost):
    def _send(self, mfrom, mto, messageText, immediate=False):
        print("<sent mail from {0} to {1}>".format(mfrom, mto))  # noqa: T003
        if hasattr(messageText, "encode"):
            # It is text instead of bytes.
            messageText = messageText.encode("utf-8")
        self.msgtext = messageText
        self.msg = message_from_bytes(messageText.lstrip())


class Fixture(PloneSandboxLayer):

    defaultBases = (PLONE_APP_CONTENTTYPES_FIXTURE,)

    def setUpZope(self, app, configurationContext):
        from plone.protect import auto  # noqa

        auto.CSRF_DISABLED = True
        import collective.easyform
        import plone.restapi

        self.loadZCML(package=collective.easyform)
        self.loadZCML(package=plone.restapi)
        try:
            import plone.formwidget.recaptcha

            self.loadZCML(package=plone.formwidget.recaptcha)

        except ImportError:
            pass
        try:
            import plone.formwidget.hcaptcha

            self.loadZCML(package=plone.formwidget.hcaptcha)
        except ImportError:
            pass

        try:
            import collective.z3cform.norobots

            self.loadZCML(package=collective.z3cform.norobots)
        except ImportError:
            pass

        # set default publisher encoding for Plone 5.1
        # this is set in zope.conf via plone.recipe.zope2instance but for
        # testbrowser in doctests we have to manually set it here
        # see https://github.com/collective/collective.easyform/pull/139
        if api.env.plone_version() < "5.2":
            from Zope2.Startup.datatypes import default_zpublisher_encoding

            default_zpublisher_encoding("utf-8")

    def setUpPloneSite(self, portal):
        # Install the collective.easyform product
        self.applyProfile(portal, "collective.easyform:default")
        try:
            self.applyProfile(portal, "plone.formwidget.recaptcha:default")
            self.applyProfile(portal, "plone.formwidget.hcaptcha:default")

        except KeyError:
            pass
        setRoles(portal, TEST_USER_ID, ["Manager"])
        portal.manage_changeProperties(email_from_address="mdummy@address.com")
        portal.MailHost = mailhost = MailHostMock()
        sm = getSiteManager(context=portal)
        sm.unregisterUtility(provided=IMailHost)
        sm.registerUtility(mailhost, provided=IMailHost)


FIXTURE = Fixture()
INTEGRATION_TESTING = IntegrationTesting(
    bases=(FIXTURE,), name="collective.easyform:Integration"
)
FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(FIXTURE,), name="collective.easyform:Functional"
)
ACCEPTANCE_TESTING = FunctionalTesting(
    bases=(FIXTURE, REMOTE_LIBRARY_BUNDLE_FIXTURE, WSGI_SERVER_FIXTURE),
    name="collective.easyform:Acceptance",
)


class EasyFormTestCase(TestCase):

    layer = INTEGRATION_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.portal.invokeFactory("Folder", "test-folder")
        self.folder = self.portal["test-folder"]
        self.afterSetUp()

    def afterSetUp(self):
        pass


class EasyFormFunctionalTestCase(TestCase):

    layer = FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.portal.invokeFactory("Folder", "test-folder")
        self.folder = self.portal["test-folder"]
        self.folder.invokeFactory("EasyForm", "ff1")
        self.ff1 = getattr(self.folder, "ff1")



# -*- coding: utf-8 -*-
#
# Integration tests specific to save-data adapter.
#

from AccessControl.unauthorized import Unauthorized
from collective.easyform.api import get_actions
from collective.easyform.api import get_schema
from collective.easyform.interfaces import ISaveData
from collective.easyform.tests import base
from os.path import dirname
from os.path import join
from plone import api
from plone.app.testing import login
from plone.app.testing import logout
from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.testing.zope import Browser
from six import BytesIO
from six.moves import zip
from transaction import commit
from zExceptions.unauthorized import Unauthorized as zUnauthorized
from ZPublisher.HTTPRequest import HTTPRequest
from ZPublisher.HTTPResponse import HTTPResponse

import plone.protect
import sys
import transaction
import unittest

try:
    from openpyxl import load_workbook
    HAS_OPENPYXL = True
except ImportError:
    HAS_OPENPYXL = False


def FakeRequest(method="GET", add_auth=False, **kwargs):
    environ = {}
    environ.setdefault("SERVER_NAME", "foo")
    environ.setdefault("SERVER_PORT", "80")
    environ.setdefault("REQUEST_METHOD", method)
    if api.env.plone_version() < "5.2":
        # manually set stdout for Plone < 5.2
        request = HTTPRequest(sys.stdin, environ, HTTPResponse(stdout=BytesIO()))
    else:
        request = HTTPRequest(sys.stdin, environ, HTTPResponse())
    request.form = kwargs
    if add_auth:
        request.form["_authenticator"] = plone.protect.createToken()
    return request


class BaseSaveData(base.EasyFormTestCase):

    """test save data adapter"""

    def setUp(self):
        base.EasyFormTestCase.setUp(self)
        self.folder.invokeFactory("EasyForm", "ff1")
        self.ff1 = getattr(self.folder, "ff1")

    def createSaver(self):
        """Creates FormCustomScript object"""
        # 1. Create custom script adapter in the form folder
        self.portal.REQUEST["form.widgets.title"] = u"Saver"
        self.portal.REQUEST["form.widgets.__name__"] = u"saver"
        self.portal.REQUEST["form.widgets.description"] = u""
        self.portal.REQUEST["form.widgets.factory"] = ["Save Data"]
        self.portal.REQUEST["form.buttons.add"] = u"Add"
        view = self.ff1.restrictedTraverse("actions/@@add-action")
        view.update()
        form = view.form_instance
        data, errors = form.extractData()
        self.assertEqual(len(errors), 0)

        # 2. Check that creation succeeded
        actions = get_actions(self.ff1)
        self.assertTrue("saver" in actions)


class SaveDataTestCase(BaseSaveData):

    def testSavedDataView(self):
        """test saved data view"""

        self.createSaver()

        view = self.ff1.restrictedTraverse("saveddata")
        self.assertEqual(list(view.items()), [("saver", u"Saver")])

        # as the owner, TEST_USER_ID can still see the saved data
        setRoles(self.portal, TEST_USER_ID, [])
        view = self.ff1.restrictedTraverse("saveddata")
        self.assertIsNotNone(view)
        logout()

        # other editors / reviewers / ... don't have access to the saved data
        api.user.create(email="test@plone.org", username="test")
        setRoles(
            self.portal, "test", roles=["Reader", "Contributor", "Editor", "Reviewer"]
        )
        login(self.portal, "test")
        with self.assertRaises(Unauthorized):
            view = self.ff1.restrictedTraverse("saveddata")

    def testSaverDataFormExtraData(self):
        """test saver data form extra data"""

        self.createSaver()

        self.assertTrue("saver" in get_actions(self.ff1))
        saver = get_actions(self.ff1)["saver"]
        self.assertEqual(saver.itemsSaved(), 0)
        request = FakeRequest(
            topic="test subject", replyto="test@test.org", comments="test comments"
        )
        saver.ExtraData = ("dt",)
        saver.onSuccess(request.form, request)

        view = self.ff1.restrictedTraverse("@@actions")
        view = view.publishTraverse(view.request, "saver")
        view = view.publishTraverse(view.request, "data")
        view.update()
        form = view.form_instance
        message = form.description()
        self.assertEqual(message.mapping, {"items": 1})
        item = form.get_items()[0]
        self.assertEqual(item[1]["id"], item[0])
        self.assertEqual(item[1]["topic"], "test subject")
        self.assertEqual(item[1]["replyto"], "test@test.org")
        self.assertEqual(item[1]["comments"], "test comments")
        self.assertTrue(" " in item[1]["dt"])

    def testSaverDataFormShowFields(self):
        """test saver data form show fields"""

        self.createSaver()

        self.assertTrue("saver" in get_actions(self.ff1))
        saver = get_actions(self.ff1)["saver"]
        self.assertEqual(saver.itemsSaved(), 0)
        request = FakeRequest(
            topic="test subject", replyto="test@test.org", comments="test comments"
        )
        saver.showFields = ("topic", "comments")
        saver.onSuccess(request.form, request)

        view = self.ff1.restrictedTraverse("@@actions")
        view = view.publishTraverse(view.request, "saver")
        view = view.publishTraverse(view.request, "data")
        view.update()
        form = view.form_instance
        message = form.description()
        self.assertEqual(message.mapping, {"items": 1})
        item = form.get_items()[0]
        self.assertEqual(item[1]["id"], item[0])
        self.assertEqual(item[1]["topic"], "test subject")
        self.assertEqual(item[1]["comments"], "test comments")
        self.assertTrue("replyto" not in item[1])

    def testSaverDataFormOneItem(self):
        """test saver data form one item"""

        self.createSaver()

        self.assertTrue("saver" in get_actions(self.ff1))
        saver = get_actions(self.ff1)["saver"]
        self.assertEqual(saver.itemsSaved(), 0)
        request = FakeRequest(
            topic="test subject", replyto="test@test.org", comments="test comments"
        )
        saver.onSuccess(request.form, request)

        view = self.ff1.restrictedTraverse("@@actions")
        view = view.publishTraverse(view.request, "saver")
        view = view.publishTraverse(view.request, "data")
        view.update()
        form = view.form_instance
        message = form.description()
        self.assertEqual(message.mapping, {"items": 1})
        item = form.get_items()[0]
        self.assertEqual(item[1]["id"], item[0])
        self.assertEqual(item[1]["topic"], "test subject")
        self.assertEqual(item[1]["replyto"], "test@test.org")
        self.assertEqual(item[1]["comments"], "test comments")

    def testSaverDataForm(self):
        """test saver data form"""

        self.createSaver()

        view = self.ff1.restrictedTraverse("@@actions")
        view = view.publishTraverse(view.request, "saver")
        view = view.publishTraverse(view.request, "data")
        view.update()
        form = view.form_instance
        message = form.description()
        self.assertEqual(message.mapping, {"items": 0})
        self.assertEqual([i for i in form.get_items()], [])

    def testSaver(self):
        """test save data adapter action"""

        self.createSaver()

        self.assertTrue("saver" in get_actions(self.ff1))
        saver = get_actions(self.ff1)["saver"]

        self.assertEqual(saver.itemsSaved(), 0)

        res = saver.getSavedFormInputForEdit()
        self.assertEqual(res, "")

        request = FakeRequest(
            add_auth=True,
            method="POST",
            topic="test subject",
            replyto="test@test.org",
            comments="test comments",
        )
        saver.onSuccess(request.form, request)

        self.assertEqual(saver.itemsSaved(), 1)

        res = saver.getSavedFormInputForEdit()
        self.assertEqual(res.strip(), "test@test.org,test subject,test comments")

    def testSaverExtraData(self):
        """test save data adapter action"""

        self.createSaver()

        self.assertTrue("saver" in get_actions(self.ff1))
        saver = get_actions(self.ff1)["saver"]
        self.assertEqual(saver.itemsSaved(), 0)

        request = FakeRequest(
            add_auth=True,
            method="POST",
            topic="test subject",
            replyto="test@test.org",
            comments="test comments",
        )
        saver.ExtraData = ("dt",)
        saver.onSuccess(request.form, request)

        self.assertEqual(saver.itemsSaved(), 1)
        self.assertTrue("dt" in saver.getSavedFormInput()[0])

    def testSaverDownload(self):
        """test save data"""

        self.createSaver()

        self.assertTrue("saver" in get_actions(self.ff1))
        saver = get_actions(self.ff1)["saver"]
        self.assertEqual(saver.itemsSaved(), 0)

        request = FakeRequest(
            add_auth=True,
            method="POST",
            topic="test subject",
            replyto="test@test.org",
            comments="test comments",
        )
        saver.onSuccess(request.form, request)

        self.assertEqual(saver.itemsSaved(), 1)
        saver.download(request.response, delimiter=",")
        res = request.response.stdout.getvalue().decode("utf-8")
        self.assertTrue("Content-Type: text/comma-separated-values" in res)
        self.assertTrue('Content-Disposition: attachment; filename="saver.csv"' in res)
        self.assertTrue(saver.getSavedFormInputForEdit() in res)

    def testSaverDownloadTSV(self):
        """test save data"""

        self.createSaver()

        self.assertTrue("saver" in get_actions(self.ff1))
        saver = get_actions(self.ff1)["saver"]
        self.assertEqual(saver.itemsSaved(), 0)

        request = FakeRequest(
            add_auth=True,
            method="POST",
            topic="test subject",
            replyto="test@test.org",
            comments="test comments",
        )
        saver.onSuccess(request.form, request)

        self.assertEqual(saver.itemsSaved(), 1)
        saver.DownloadFormat = "tsv"
        saver.download(request.response)
        res = request.response.stdout.getvalue().decode("utf-8")
        self.assertTrue("Content-Type: text/tab-separated-values" in res)
        self.assertTrue('Content-Disposition: attachment; filename="saver.tsv"' in res)
        self.assertTrue(saver.getSavedFormInputForEdit(delimiter="\t") in res)

    @unittest.skipUnless(HAS_OPENPYXL, "Requires openpyxl")
    def testSaverDownloadXLSX(self):
        """test save data"""

        with open(join(dirname(__file__), "fixtures", "fieldset_multiple_choice.xml")) as f:
            self.ff1.fields_model = f.read()

        self.createSaver()

        self.assertTrue("saver" in get_actions(self.ff1))
        saver = get_actions(self.ff1)["saver"]
        self.assertEqual(saver.itemsSaved(), 0)

        request = FakeRequest(
            add_auth=True,
            method="POST",
            topic="test subject",
            replyto="test@test.org",
            comments="test comments",
            multiplechoice=["Red", "Blue"]
        )
        saver.onSuccess(request.form, request)

        self.assertEqual(saver.itemsSaved(), 1)
        saver.DownloadFormat = "xlsx"
        saver.download(request.response)
        res = request.response.stdout
        self.assertEqual(
            request.response.headers["content-type"],
            "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
        )
        self.assertEqual(
            request.response.headers["content-disposition"],
            'attachment; filename="saver.xlsx"',
        )

        wb = load_workbook(res)
        wb.active
        ws = wb.active
        rows = list(ws.rows)

        self.assertEqual(len(rows), 2)  # Row 1 is the header row
        self.assertEqual(rows[1][0].value, "test@test.org")
        self.assertEqual(rows[1][1].value, "test subject")
        self.assertEqual(rows[1][2].value, "test comments")
        self.assertEqual(rows[1][3].value, "Red|Blue")

    def testSaverDownloadWithTitles(self):
        """test save data"""

        self.createSaver()

        self.assertTrue("saver" in get_actions(self.ff1))
        saver = get_actions(self.ff1)["saver"]
        self.assertEqual(saver.itemsSaved(), 0)

        request = FakeRequest(
            add_auth=True,
            method="POST",
            topic="test subject",
            replyto="test@test.org",
            comments="test comments",
        )
        saver.onSuccess(request.form, request)

        self.assertEqual(saver.itemsSaved(), 1)
        saver.UseColumnNames = True
        saver.download(request.response, delimiter=",")
        res = request.response.stdout.getvalue().decode("utf-8")
        self.assertTrue("Content-Type: text/comma-separated-values" in res)
        self.assertTrue('Content-Disposition: attachment; filename="saver.csv"' in res)
        self.assertTrue(saver.getSavedFormInputForEdit(header=True) in res)

    def testSaverDownloadExtraData(self):
        """test save data"""

        self.createSaver()

        self.assertTrue("saver" in get_actions(self.ff1))
        saver = get_actions(self.ff1)["saver"]
        self.assertEqual(saver.itemsSaved(), 0)

        request = FakeRequest(
            add_auth=True,
            method="POST",
            topic="test subject",
            replyto="test@test.org",
            comments="test comments",
        )
        saver.ExtraData = ("dt", "HTTP_USER_AGENT")
        saver.onSuccess(request.form, request)

        self.assertEqual(saver.itemsSaved(), 1)
        saver.download(request.response, delimiter=",")
        res = request.response.stdout.getvalue().decode("utf-8")
        self.assertTrue("Content-Type: text/comma-separated-values" in res)
        self.assertTrue('Content-Disposition: attachment; filename="saver.csv"' in res)
        self.assertTrue(saver.getSavedFormInputForEdit() in res)

    def testSaverSavedFormInput(self):
        """test save data adapter action and direct access to
        SavedFormInput"""

        self.createSaver()

        self.assertTrue("saver" in get_actions(self.ff1))
        saver = get_actions(self.ff1)["saver"]

        request = FakeRequest(
            add_auth=True,
            method="POST",
            topic="test subject",
            replyto="test@test.org",
            comments="test comments",
        )
        saver.onSuccess(request.form, request)

        self.assertEqual(saver.itemsSaved(), 1)
        row = next(iter(saver.getSavedFormInput()))
        self.assertEqual(len(row), 4)

        request = FakeRequest(
            add_auth=True,
            method="POST",
            topic="test subject",
            replyto="test@test.org",
            comments="test comments",
        )
        saver.onSuccess(request.form, request)
        self.assertEqual(saver.itemsSaved(), 2)

        saver.clearSavedFormInput()
        self.assertEqual(saver.itemsSaved(), 0)

    def testSetSavedFormInput(self):
        """test setSavedFormInput functionality"""

        # set up saver
        self.createSaver()
        self.assertTrue("saver" in get_actions(self.ff1))
        saver = get_actions(self.ff1)["saver"]
        self.assertEqual(saver.getSavedFormInputForEdit(), "")

        # save a row
        fields = list(get_schema(self.ff1))
        saver.addDataRow(dict(list(zip(fields, ["one", "two", "three"]))))
        self.assertEqual(saver.itemsSaved(), 1)
        items = saver.getSavedFormInputItems()
        self.assertEqual(
            items[0][1],
            dict(list(zip(["id"] + fields, [items[0][0], "one", "two", "three"]))),
        )
        for number in ["one", "two", "three"]:
            self.assertIn(number, saver.getSavedFormInputForEdit())

        # save a couple of \n-delimited rows - \n eol
        saver.addDataRow(dict(list(zip(fields, ["four", "five", "six"]))))
        self.assertEqual(saver.itemsSaved(), 2)
        items = saver.getSavedFormInputItems()
        self.assertEqual(
            items[0][1],
            dict(list(zip(["id"] + fields, [items[0][0], "one", "two", "three"]))),
        )
        self.assertEqual(
            items[1][1],
            dict(list(zip(["id"] + fields, [items[1][0], "four", "five", "six"]))),
        )

        # order can change in py2
        for number in ["one", "two", "three", "four", "five", "six"]:
            self.assertIn(number, saver.getSavedFormInputForEdit())

        # save empty string
        saver.clearSavedFormInput()
        self.assertEqual(saver.itemsSaved(), 0)
        self.assertEqual(saver.getSavedFormInputForEdit(), "")

    def testEditSavedFormInput(self):
        """test manage_saveData functionality"""

        # set up saver
        self.createSaver()
        self.assertTrue("saver" in get_actions(self.ff1))
        saver = get_actions(self.ff1)["saver"]

        # save a row
        fields = list(get_schema(self.ff1))
        # saver.savedFormInput = 'one,two,three'
        saver.addDataRow(dict(list(zip(fields, ["one", "two", "three"]))))
        self.assertEqual(saver.itemsSaved(), 1)
        items = saver.getSavedFormInputItems()
        self.assertEqual(
            items[0][1],
            dict(list(zip(["id"] + fields, [items[0][0], "one", "two", "three"]))),
        )

    def testDeleteSavedFormInput(self):
        """test manage_deleteData functionality"""

        # set up saver
        self.createSaver()
        self.assertTrue("saver" in get_actions(self.ff1))
        saver = get_actions(self.ff1)["saver"]

        # save a few rows
        fields = list(get_schema(self.ff1))
        saver.addDataRow(dict(list(zip(fields, ["one", "two", "three"]))))
        saver.addDataRow(dict(list(zip(fields, ["four", "five", "six"]))))
        saver.addDataRow(dict(list(zip(fields, ["seven", "eight", "nine"]))))
        self.assertEqual(saver.itemsSaved(), 3)

        # saver.manage_deleteData(saver._storage.keys()[1])
        saver.delDataRow(saver.getSavedFormInputItems()[1][0])
        self.assertEqual(saver.itemsSaved(), 2)
        items = saver.getSavedFormInputItems()
        self.assertEqual(
            items[0][1],
            dict(list(zip(["id"] + fields, [items[0][0], "one", "two", "three"]))),
        )
        self.assertEqual(
            items[1][1],
            dict(list(zip(["id"] + fields, [items[1][0], "seven", "eight", "nine"]))),
        )

    def testSaverInputAsDictionaries(self):
        """test save data adapter's InputAsDictionaries"""

        self.createSaver()

        self.assertTrue("saver" in get_actions(self.ff1))
        saver = get_actions(self.ff1)["saver"]

        # self.ff1.setActionAdapter(('saver',))

        # self.assertEqual(saver.inputAsDictionaries,
        # saver.InputAsDictionaries)

        self.assertEqual(saver.itemsSaved(), 0)

        request = FakeRequest(
            add_auth=True,
            method="POST",
            topic="test subject",
            replyto="test@test.org",
            comments="test comments",
        )
        saver.onSuccess(request.form, request)
        # errors = self.ff1.fgvalidate(REQUEST=request)
        # self.assertEqual(errors, {})

        self.assertEqual(saver.itemsSaved(), 1)

        iad = saver.getSavedFormInput()
        row = next(iter(iad))
        self.assertEqual(len(row), 4)
        self.assertEqual(row["topic"], "test subject")

    def testSaverColumnNames(self):
        """test save data adapter's getColumnNames function"""

        self.createSaver()

        self.assertTrue("saver" in get_actions(self.ff1))
        saver = get_actions(self.ff1)["saver"]

        cn = saver.getColumnNames()
        self.assertTrue(len(cn) == 3)
        self.assertTrue(cn[0] == "replyto")
        self.assertTrue(cn[1] == "topic")
        self.assertTrue(cn[2] == "comments")

        # Use selective field saving
        saver.showFields = ("topic", "comments")
        cn = saver.getColumnNames()
        self.assertTrue(len(cn) == 2)
        self.assertTrue(cn[0] == "topic")
        self.assertTrue(cn[1] == "comments")
        saver.showFields = tuple()

        # Add an extra column
        saver.ExtraData = ("dt",)
        cn = saver.getColumnNames()
        self.assertTrue(len(cn) == 4)
        self.assertTrue(cn[3] == "dt")

    def testSaverColumnTitles(self):
        """test save data adapter's getColumnTitles function"""

        self.createSaver()

        self.assertTrue("saver" in get_actions(self.ff1))
        saver = get_actions(self.ff1)["saver"]

        cn = saver.getColumnTitles()
        self.assertTrue(len(cn) == 3)
        self.assertTrue(cn[0] == "Your E-Mail Address")
        self.assertTrue(cn[1] == "Subject")
        self.assertTrue(cn[2] == "Comments")

        # Add an extra column
        saver.ExtraData = ("dt",)
        cn = saver.getColumnTitles()
        self.assertTrue(len(cn) == 4)
        self.assertTrue(cn[3] == "Posting Date/Time")

    def testSaverSelectiveFieldSaving(self):
        """Test selective inclusion of fields in the data"""

        self.createSaver()

        self.assertTrue("saver" in get_actions(self.ff1))
        saver = get_actions(self.ff1)["saver"]
        saver.showFields = ("topic", "comments")

        request = FakeRequest(
            add_auth=True,
            method="POST",
            topic="test subject",
            replyto="test@test.org",
            comments="test comments",
        )
        saver.onSuccess(request.form, request)

        self.assertEqual(saver.itemsSaved(), 1)
        row = next(iter(saver.getSavedFormInput()))
        self.assertEqual(len(row), 3)
        self.assertEqual(row["topic"], "test subject")
        self.assertEqual(row["comments"], "test comments")

    def testGetSaveDataAdaptersView(self):
        """test the @@get_save_data_adapters view"""

        self.createSaver()

        view = self.ff1.restrictedTraverse("@@get_save_data_adapters")
        results = view()
        self.assertEqual(len(results), 1)
        saver = results[0]
        self.assertTrue(ISaveData.providedBy(saver))


class SaverIntegrationTestCase(base.EasyFormFunctionalTestCase):
    def setUp(self):
        base.EasyFormFunctionalTestCase.setUp(self)
        self.portal = self.layer["portal"]
        self.portal_url = self.layer["portal"].absolute_url()
        self.browser = Browser(self.layer["app"])
        self.browser.handleErrors = False
        self.browser.addHeader(
            "Authorization", "Basic " + SITE_OWNER_NAME + ":" + SITE_OWNER_PASSWORD
        )
        self.createSaver()

    def createSaver(self):
        """Creates FormCustomScript object"""
        # 1. Create custom script adapter in the form folder
        self.portal.REQUEST["form.widgets.title"] = u"Saver"
        self.portal.REQUEST["form.widgets.__name__"] = u"saver"
        self.portal.REQUEST["form.widgets.description"] = u""
        self.portal.REQUEST["form.widgets.factory"] = ["Save Data"]
        self.portal.REQUEST["form.buttons.add"] = u"Add"
        view = self.ff1.restrictedTraverse("actions/@@add-action")
        view.update()
        commit()
        form = view.form_instance
        data, errors = form.extractData()
        self.assertEqual(len(errors), 0)

        # 2. Check that creation succeeded
        actions = get_actions(self.ff1)
        self.assertTrue("saver" in actions)

    def test_download_saveddata_view(self):
        self.browser.open(self.portal_url + "/test-folder/ff1/@@actions/saver/@@data")
        self.assertTrue("Saved Data" in self.browser.contents)
        self.assertTrue(
            "viewpermission-collective-easyform-download-saved-input"
            in self.browser.contents
        )

        # editors / reviewers / ... don't have access to the saved data
        api.user.create(email="test@plone.org", username="test", password="password")
        setRoles(
            self.portal, "test", roles=["Reader", "Contributor", "Editor", "Reviewer"]
        )
        self.browser.addHeader("Authorization", "Basic " + "test" + ":" + "password")
        transaction.commit()
        with self.assertRaises(zUnauthorized):
            self.browser.open(
                self.portal_url + "/test-folder/ff1/@@actions/saver/@@data"
            )

    def test_download_saveddata_suggests_csv_delimiter_default_registry_value(self):
        self.browser.open(self.portal_url + "/test-folder/ff1/@@actions/saver/@@data")
        self.assertTrue("Saved Data" in self.browser.contents)
        input = self.browser.getControl("CSV delimiter")
        self.assertEqual(input.value, ",")

    def test_download_saveddata_suggests_csv_delimiter_updated_registry_value(self):
        # 1. set custom CSV delimiter in registry
        self.browser.open(self.portal_url + "/@@easyform-controlpanel")
        input = self.browser.getControl(label="CSV delimiter")
        input.value = ";"
        self.browser.getControl("Save").click()
        # 2. tests that custom value is used
        self.browser.open(self.portal_url + "/test-folder/ff1/@@actions/saver/@@data")
        input = self.browser.getControl("CSV delimiter")
        self.assertEqual(input.value, ";")

    def test_download_saveddata_csv_delimiter_required(self):
        self.browser.open(self.portal_url + "/test-folder/ff1/@@actions/saver/@@data")
        self.assertTrue("Saved Data" in self.browser.contents)
        input = self.browser.getControl("CSV delimiter")
        input.value = ""
        self.browser.getControl(name="form.buttons.download").click()
        self.assertEqual(
            self.browser.url,
            "http://nohost/plone/test-folder/ff1/@@actions/saver/@@data",
        )
        self.assertTrue("CSV delimiter is required." in self.browser.contents)

    def test_download_saveddata_csv_delimiter_from_form(self):
        self.browser.open(
            self.portal_url + "/test-folder/ff1/@@actions/saver/@@data"
        )
        self.assertTrue("Saved Data" in self.browser.contents)
        self.browser.getControl(name="form.buttons.download").click()
        self.assertEqual(
            self.browser.contents, "Your E-Mail Address,Subject,Comments\r\n"
        )
        self.browser.open(
            self.portal_url + "/test-folder/ff1/@@actions/saver/@@data"
        )
        self.assertTrue("Saved Data" in self.browser.contents)
        input = self.browser.getControl("CSV delimiter")
        input.value = ";"
        self.browser.getControl(name="form.buttons.download").click()
        self.assertEqual(
            self.browser.contents, "Your E-Mail Address;Subject;Comments\r\n"
        )

    # this test depends on internals of zope.testbrowser controls
    def test_download_saveddata_suggests_csv_delimiter_defines_maxlength(self):
        self.browser.open(
            self.portal_url + "/test-folder/ff1/@@actions/saver/@@data"
        )
        self.assertTrue("Saved Data" in self.browser.contents)
        input = self.browser.getControl("CSV delimiter")
        self.assertTrue(input._elem.has_attr("maxlength"))
        self.assertEqual(input._elem.get("maxlength"), u"1")



# -*- coding: utf-8 -*-
from collective.easyform import easyformMessageFactory as _
from collective.easyform.interfaces import IFieldValidator
from plone import api
from Products.CMFPlone.RegistrationTool import EmailAddressInvalid
from Products.validation import validation
from Products.validation.validators.BaseValidators import baseValidators
from zope.component import provideUtility

import six


BAD_SIGNS = frozenset(["<a ", "www.", "http:", ".com", "https:"])


def isValidEmail(value, **kwargs):
    """Check for the user email address."""
    if value is None:
        return
    reg_tool = api.portal.get_tool("portal_registration")
    if not (value and reg_tool.isValidEmail(value)):
        raise EmailAddressInvalid


def isCommaSeparatedEmails(value, **kwargs):
    """Check for one or more E-Mail Addresses separated by commas."""
    if value is None:
        # Let the system for required take care of None values
        return
    reg_tool = api.portal.get_tool("portal_registration")
    for v in value.split(","):
        if not reg_tool.isValidEmail(v.strip()):
            return _(
                u"Must be a valid list of email addresses " u"(separated by commas)."
            )


def isChecked(value, **kwargs):
    if not (
        (isinstance(value, bool) and value)
        or (isinstance(value, six.string_types) and value == "1")
    ):
        return _(u"Must be checked.")


def isUnchecked(value, **kwargs):
    if not isChecked(value):
        return _(u"Must be unchecked.")


def isNotLinkSpam(value, **kwargs):
    if not value:
        return  # No value can't be SPAM
    # validation is optional and configured on the field
    value = value.lower()
    for s in BAD_SIGNS:
        if s in value:
            return _("Links are not allowed.")


# Base validators
def update_validators():
    if validation and baseValidators:

        def method(name):
            def validate(value, **kwargs):
                if value is None:
                    # Let the system for required take care of None values
                    return
                if six.PY2 and isinstance(value, six.text_type):
                    value = value.encode("utf-8")
                res = validation(name, value, **kwargs)
                if res != 1:
                    return res

            return validate

        for validator in baseValidators:
            provideUtility(
                method(validator.name), provides=IFieldValidator, name=validator.name
            )


update_validators()



# -*- coding: utf-8 -*-

from collective.easyform import easyformMessageFactory as _
from collective.easyform.interfaces import IEasyFormImportFormSchema
from datetime import datetime
from plone.z3cform import layout
from Products.Five import BrowserView
from Products.GenericSetup.context import TarballExportContext
from Products.GenericSetup.context import TarballImportContext
from Products.GenericSetup.interfaces import IFilesystemExporter
from Products.GenericSetup.interfaces import IFilesystemImporter
from Products.statusmessages.interfaces import IStatusMessage
from z3c.form import button
from z3c.form import field
from z3c.form import form
from zope.component import getMultiAdapter


class EasyFormExportView(BrowserView):
    """See ..interfaces.exportimport.IEasyFormExportView."""

    def __call__(self):
        """See ..interfaces.exportimport.IEasyFormExportView.__call__"""
        ctx = TarballExportContext(self.context)
        response = self.request.RESPONSE
        disposition = 'attachment; filename="{0}-{1:{2}}.tar.gz"'.format(
            self.context.getId(), datetime.now(), "%Y%m%d%H%M%S"
        )

        response.setHeader("Content-type", "application/x-gzip")
        response.setHeader("Content-disposition", disposition)

        # export the structure treating the current form as our root context
        IFilesystemExporter(self.context).export(ctx, "structure", True)

        return ctx.getArchive()


class EasyFormImportForm(form.Form):
    """The form class for importing of exported easyforms."""

    fields = field.Fields(IEasyFormImportFormSchema)
    ignoreContext = True
    ignoreReadonly = True

    @button.buttonAndHandler(_(u"import"), name="import")
    def handleImport(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return

        ctx = TarballImportContext(self.context, data["upload"])
        IFilesystemImporter(self.context).import_(ctx, "structure", True)

        self.status = _(u"Form imported.")
        IStatusMessage(self.request).addStatusMessage(self.status, type="info")

        url = getMultiAdapter((self.context, self.request), name="absolute_url")()
        self.request.response.redirect(url)


EasyFormImportView = layout.wrap_form(EasyFormImportForm)



# -*- coding: utf-8 -*-
from ast import literal_eval
from collective.easyform.api import get_actions
from collective.easyform.api import get_schema
from collective.easyform.interfaces import ISaveData
from DateTime import DateTime
from plone.namedfile.interfaces import INamedBlobFileField
from zope.schema.interfaces import IDate
from zope.schema.interfaces import IDatetime
from zope.schema.interfaces import IFromUnicode
from zope.schema.interfaces import ISet

import logging


logger = logging.getLogger("collective.easyform.migration")


def migrate_saved_data(ploneformgen, easyform):
    for data_adapter in ploneformgen.objectValues("FormSaveDataAdapter"):
        actions = get_actions(easyform)
        action = actions.get(data_adapter.getId())
        schema = get_schema(easyform)
        if ISaveData.providedBy(action):
            cols = data_adapter.getColumnNames()
            for idx, row in enumerate(data_adapter.getSavedFormInput()):
                if len(row) != len(cols):
                    logger.warning(
                        "Number of columns does not match. Skipping row %s in "
                        "data adapter %s/%s",
                        idx,
                        "/".join(easyform.getPhysicalPath()),
                        data_adapter.getId(),
                    )
                    continue
                data = {}
                for key, value in zip(cols, row):
                    field = schema.get(key)
                    value = value.decode("utf8")
                    if IFromUnicode.providedBy(field):
                        value = field.fromUnicode(value)
                    elif IDatetime.providedBy(field) and value:
                        value = DateTime(value).asdatetime()
                    elif IDate.providedBy(field) and value:
                        value = DateTime(value).asdatetime().date()
                    elif ISet.providedBy(field):
                        try:
                            value = set(literal_eval(value))
                        except ValueError:
                            pass
                    elif INamedBlobFileField.providedBy(field):
                        value = None
                    data[key] = value
                action.addDataRow(data)



# -*- coding: utf-8 -*-
"""

    Unit test for EasyForm custom scripts

    Copyright 2006 Red Innovation http://www.redinnovation.com

"""
from AccessControl import ClassSecurityInfo
from AccessControl import Unauthorized
from AccessControl.class_init import InitializeClass
from collective.easyform.api import get_actions
from collective.easyform.tests import base
from plone.app.testing import logout
from Products.CMFCore import permissions

import six


test_script = """
## Python Script
##bind container=container
##bind context=context
##bind subpath=traverse_subpath
##parameters=fields, easyform, request
##title=Succesfully working script
##

from Products.CMFCore.utils import getToolByName

assert fields['test_field'] == '123'
return 'foo'
"""

bad_parameters_script = """
## Python Script
##bind container=container
##bind context=context
##bind subpath=traverse_subpath
##parameters=
##title=
##

return 'foo'
"""

syntax_error_script = """
## Python Script
##bind container=container
##bind context=context
##bind subpath=traverse_subpath
##parameters=fields, easyform, request
##title=
##
if:
return asdfaf
"""

runtime_error_script = """
## Python Script
##bind container=container
##bind context=context
##bind subpath=traverse_subpath
##parameters=fields, easyform
##title=
##
return 'asdfaf' + 1
"""

security_script = """
## Python Script
##bind container=container
##bind context=context
##bind subpath=traverse_subpath
##parameters=fields, easyform, request
##title=Script needing special privileges
##

from Products.CMFCore.utils import getToolByName

portal_url = getToolByName(context, 'portal_url')
portal = portal_url.getPortalObject()

print(portal)

# Try set left_slots
portal.manage_addProperty('foo', ['foo'], 'lines')
"""

proxied_script = """
## Python Script
##bind container=container
##bind context=context
##bind subpath=traverse_subpath
##parameters=fields, easyform, request
##title=Script needing special privileges
##

# Should raise Unauthorized
return request.fooProtected()
"""

return_error_script = """
## Python Script
##bind container=container
##bind context=context
##bind subpath=traverse_subpath
##parameters=fields, easyform, request
##title=
##
return {'comments': 'Please enter more text'}
"""

default_params_script = """
fields
easyform
request
"""


class FakeRequest(dict):
    def __init__(self, **kwargs):
        self.form = kwargs


class SecureFakeRequest(dict):

    security = ClassSecurityInfo()

    def __init__(self, **kwargs):
        self.form = kwargs

    @security.protected(permissions.ManagePortal)
    def fooProtected(self):
        """Only manager can access this"""
        return "foo"


InitializeClass(SecureFakeRequest)


class TestCustomScript(base.EasyFormTestCase):

    """Test FormCustomScriptAdapter functionality in EasyForm"""

    def afterSetUp(self):
        self.request = self.layer["request"]

        self.folder.invokeFactory("EasyForm", "ff1")
        self.ff1 = getattr(self.folder, "ff1")
        self.ff1.CSRFProtection = False
        self.request["form.widgets.title"] = u"Test field"
        self.request["form.widgets.__name__"] = u"test_field"
        self.request["form.widgets.description"] = u"foobar"
        self.request["form.widgets.factory"] = ["label_textline_field"]
        self.request["form.widgets.required"] = ["selected"]
        self.request["form.widgets.fieldset_id"] = "0"
        self.request["form.buttons.add"] = u"Add"
        view = self.ff1.restrictedTraverse("fields/@@add-field")
        view.update()
        form = view.form_instance
        data, errors = form.extractData()
        self.assertEqual(len(errors), 0)

    def createScript(self):
        """Creates FormCustomScript object"""
        # 1. Create custom script adapter in the form folder
        self.request["form.widgets.title"] = u"Adapter"
        self.request["form.widgets.__name__"] = u"adapter"
        self.request["form.widgets.description"] = u""
        self.request["form.widgets.factory"] = ["Custom Script"]
        self.request["form.buttons.add"] = u"Add"
        view = self.ff1.restrictedTraverse("actions/@@add-action")
        view.update()
        form = view.form_instance
        data, errors = form.extractData()
        self.assertEqual(len(errors), 0)
        # 2. Check that creation succeeded
        actions = get_actions(self.ff1)
        self.assertTrue("adapter" in actions)

    def testSuccess(self):
        """Succesful script execution

        Creates a script, some form content,
        executes form handling.
        """

        self.createScript()

        actions = get_actions(self.ff1)
        adapter = actions["adapter"]

        # 4. Set script data
        adapter.ScriptBody = test_script

        req = FakeRequest(test_field="123")

        reply = adapter.onSuccess({}, req)
        assert reply == "foo", "Script returned:" + str(reply)

    def testRunTimeError(self):
        """Script has run-time error"""
        self.createScript()

        actions = get_actions(self.ff1)
        adapter = actions["adapter"]

        # 4. Set script data
        adapter.ScriptBody = runtime_error_script

        # Execute script
        throwed = False
        try:
            reply = adapter.onSuccess({})
        except TypeError:
            reply = None
            throwed = True

        assert throwed, "Bad script didn't throw run-time exception, got " + str(reply)
        assert reply is None

    def testSyntaxError(self):
        """Script has syntax errors

        TODO: Syntax errors are not returned in validation?
        """

        # Note: this test logs an error message; it does not indicate test
        # failure

        self.createScript()

        actions = get_actions(self.ff1)
        adapter = actions["adapter"]

        # 4. Set script data
        adapter.ScriptBody = syntax_error_script

        # Execute script
        throwed = False
        try:
            adapter.onSuccess({}, FakeRequest())
        except ValueError:
            throwed = True

        assert throwed, "Bad script didn't throw run-time exception"

    def testBadParameters(self):
        """Invalid number of script parameters"""

        self.createScript()

        actions = get_actions(self.ff1)
        adapter = actions["adapter"]

        # 4. Set script data
        adapter.ScriptBody = bad_parameters_script

        # Execute script
        throwed = False
        try:
            adapter.onSuccess([])
        except TypeError:
            throwed = True
        assert throwed, "Invalid parameters failed silently"

    def testDefaultParameters(self):
        """Test to make sure the documented parameters are available"""

        self.createScript()

        actions = get_actions(self.ff1)
        adapter = actions["adapter"]

        # 4. Set script data
        adapter.ScriptBody = default_params_script

        request = FakeRequest(
            topic="test subject", replyto="test@test.org", comments="test comments"
        )

        errors = adapter.onSuccess({}, request)
        self.assertEqual(errors, None)

    def testSecurity(self):
        """Script needing proxy role

        TODO: Why no security exceptions are raised?
        """
        self.createScript()
        logout()

        actions = get_actions(self.ff1)
        adapter = actions["adapter"]

        # 4. Set script data
        adapter.ScriptBody = security_script

        # Execute script
        throwed = False
        try:
            adapter.onSuccess({}, FakeRequest())
        except Unauthorized:
            throwed = True

        if self.portal.hasProperty("foo"):
            assert "Script executed under full priviledges"

        self.assertTrue(throwed, "Bypassed security, baaad!")

        adapter.ProxyRole = u"Manager"
        throwed = False
        try:
            adapter.onSuccess({}, FakeRequest())
        except Unauthorized:
            throwed = True

        if not self.portal.hasProperty("foo"):
            assert "Script not executed thru proxy role"
        self.assertFalse(throwed, "Unauthorized was raised!")

    def testSetProxyRole(self):
        """Exercise setProxyRole"""
        self.createScript()
        self.request["form.widgets.title"] = u"Adapter"
        self.request["form.widgets.description"] = u""
        self.request["form.widgets.ProxyRole"] = [u"Manager"]
        self.request["form.widgets.ScriptBody"] = six.text_type(proxied_script)
        self.request["form.widgets.IActionExtender.execCondition"] = u""
        self.request["form.buttons.save"] = u"Save"
        view = self.ff1.restrictedTraverse("actions")
        view = view.publishTraverse(view.request, "adapter")
        view = view.publishTraverse(view.request, "adapter")
        view.update()
        form = view.form_instance
        data, errors = form.extractData()
        self.assertEqual(len(errors), 0)
        self.request["form.widgets.title"] = u"Adapter"
        self.request["form.widgets.description"] = u""
        self.request["form.widgets.ProxyRole"] = [u"none"]
        self.request["form.widgets.ScriptBody"] = six.text_type(proxied_script)
        self.request["form.widgets.IActionExtender.execCondition"] = u""
        self.request["form.buttons.save"] = u"Save"
        view = self.ff1.restrictedTraverse("actions")
        view = view.publishTraverse(view.request, "adapter")
        view = view.publishTraverse(view.request, "adapter")
        view.update()
        form = view.form_instance
        data, errors = form.extractData()
        self.assertEqual(len(errors), 0)
        self.request["form.widgets.title"] = u"Adapter"
        self.request["form.widgets.description"] = u""
        self.request["form.widgets.ProxyRole"] = [u"bogus"]
        self.request["form.widgets.ScriptBody"] = six.text_type(proxied_script)
        self.request["form.widgets.IActionExtender.execCondition"] = u""
        self.request["form.buttons.save"] = u"Save"
        view = self.ff1.restrictedTraverse("actions")
        view = view.publishTraverse(view.request, "adapter")
        view = view.publishTraverse(view.request, "adapter")
        view.update()
        form = view.form_instance
        data, errors = form.extractData()
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].message, u"Required input is missing.")

    def testProxyRole(self):
        """Test seeing how setting proxy role affects unauthorized
        Exception
        """
        self.createScript()

        actions = get_actions(self.ff1)
        adapter = actions["adapter"]

        # 4. Set script data
        adapter.ScriptBody = proxied_script

        req = SecureFakeRequest(test_field="123")
        self.assertRaises(Unauthorized, adapter.onSuccess, {}, req)



import datetime
from datetime import date, datetime
import imp
import pdb
from dateutil import parser
import json
import logging

from zope.component import adapter
from zope.interface import implementer
from zope.interface import Interface
from zope.schema import getFieldsInOrder
from zope.schema.interfaces import ISet, IDate, IDatetime

from plone.restapi.serializer.dxcontent import SerializeToJson as DXContentToJson
from plone.restapi.deserializer.dxcontent import (
    DeserializeFromJson as DXContentFromJson,
)
from plone.restapi.deserializer import json_body
from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.interfaces import IDeserializeFromJson
from plone.app.textfield.value import RichTextValue
from plone.app.textfield.interfaces import IRichText

from collective.easyform.api import get_actions
from collective.easyform.api import get_schema
from collective.easyform.interfaces import IEasyForm
from collective.easyform.interfaces import ISaveData
from Products.CMFPlone.utils import safe_unicode


logger = logging.getLogger("collective.easyform.migration")


@implementer(ISerializeToJson)
@adapter(IEasyForm, Interface)
class SerializeToJson(DXContentToJson):
    def __call__(self, version=None, include_items=True):
        result = super(SerializeToJson, self).__call__(version, include_items)
        self.serializeSavedData(result)

        return result

    def serializeSavedData(self, result):
        storage = dict()
        actions = getFieldsInOrder(get_actions(self.context))

        AllFieldsinOrder = getFieldsInOrder(get_schema(self.context))
        included_columns_in_savedata = []
        for column, field in AllFieldsinOrder:
            if "label" not in field.__str__().lower():
                included_columns_in_savedata.append(column)
        included_columns_in_savedata.sort()

        for name, action in actions:
            if ISaveData.providedBy(action):
                serializeable = dict()
                storage[name] = serializeable
                for id, data in action.getSavedFormInputItems():
                    column_names = list(data.keys())
                    column_names.remove("id")
                    column_names.sort()
                    if column_names != included_columns_in_savedata:
                        logger.warning(
                            "Skipped Saveddata row because of mismatch witch current fields in %s",
                            self.context.absolute_url(),
                        )
                        continue
                    try:
                        for key, value in data.items():
                            data[key] = convertBeforeSerialize(value)
                        json.dumps(data)
                        serializeable[id] = data
                    except TypeError:
                        logger.exception(
                            "saved data serialization issue for {}".format(
                                self.context.absolute_url()
                            )
                        )
        if storage:
            result["savedDataStorage"] = storage


def convertBeforeSerialize(value):
    if isinstance(value, (datetime, date)):
        return value.isoformat()
    elif isinstance(value, set):
        return list(value)
    elif isinstance(value, RichTextValue):
        return safe_unicode(value.raw) #raw_encoded
    else:
        return value


@implementer(IDeserializeFromJson)
@adapter(IEasyForm, Interface)
class DeserializeFromJson(DXContentFromJson):
    def __call__(
        self,
        validate_all=False,
        data=None,
        create=False,
    ):  # noqa: ignore=C901

        if data is None:
            data = json_body(self.request)

        super(DeserializeFromJson, self).__call__(validate_all, data, create)

        self.deserializeSavedData(data)
        return self.context

    def deserializeSavedData(self, data):
        if "savedDataStorage" in data:
            storage = data["savedDataStorage"]
            actions = getFieldsInOrder(get_actions(self.context))
            schema = get_schema(self.context)

            AllFieldsinOrder = schema.namesAndDescriptions()
            included_columns_in_savedata = []
            for column, field in AllFieldsinOrder:
                if "label" not in field.__str__().lower():
                    included_columns_in_savedata.append(column)

            for name, action in actions:
                if ISaveData.providedBy(action) and name in storage:
                    savedData = storage[name]
                    for key, value in savedData.items():
                        for name in included_columns_in_savedata:  # schema.names():
                            value[name] = convertAfterDeserialize(
                                schema[name], value[name]
                            )
                        action.setDataRow(int(key), value)


def convertAfterDeserialize(field, value):
    if ISet.providedBy(field):
        return set(value)
    elif IDate.providedBy(field) or IDatetime.providedBy(field):
        return parser.parse(value)
    elif IRichText.providedBy(field):
        return RichTextValue(value)
    else:
        return value



# -*- coding: utf-8 -*-
from collective.easyform import easyformMessageFactory as _
from plone.app.registry.browser.controlpanel import ControlPanelFormWrapper
from plone.app.registry.browser.controlpanel import RegistryEditForm
from plone.registry.interfaces import IRegistry
from plone.z3cform import layout
from z3c.form.browser.checkbox import CheckBoxFieldWidget
from zope import schema
from zope.component import getUtility
from zope.interface import Interface


def getContent(self):
    return getUtility(IRegistry).forInterface(self.schema, prefix=self.schema_prefix)


class IEasyFormControlPanel(Interface):

    allowedFields = schema.List(
        title=_(u"Allowed Fields"),
        description=_(u"These Fields are available for your forms."),
        value_type=schema.Choice(
            required=False,
            vocabulary="easyform.SchemaEditorFields",
        ),
        default=[],
    )

    csv_delimiter = schema.TextLine(
        title=_(u"CSV delimiter"),
        max_length=1,
        description=_(u"Set the default delimiter for CSV download."),
        required=True,
        default=u",",
    )


class EasyFormControlPanelForm(RegistryEditForm):
    schema = IEasyFormControlPanel
    schema_prefix = "easyform"
    label = _(u"easyform Settings")

    def updateFields(self):
        super(EasyFormControlPanelForm, self).updateFields()
        self.fields["allowedFields"].widgetFactory = CheckBoxFieldWidget


EasyFormControlPanelView = layout.wrap_form(
    EasyFormControlPanelForm, ControlPanelFormWrapper
)



# -*- coding: utf-8 -*-
from .validators import isTALES
from collective.easyform import config
from collective.easyform import easyformMessageFactory as _  # NOQA
from plone.autoform import directives
from plone.schemaeditor.interfaces import ID_RE
from plone.schemaeditor.interfaces import IFieldContext
from plone.schemaeditor.interfaces import IFieldEditorExtender
from plone.schemaeditor.interfaces import ISchemaContext
from plone.supermodel.directives import fieldset
from plone.supermodel.model import Schema

import z3c.form.interfaces
import zope.interface
import zope.schema.interfaces


try:
    from plone.schemaeditor import SchemaEditorMessageFactory as __  # NOQA
except ImportError:
    from plone.schemaeditor import _ as __


MODIFY_PORTAL_CONTENT = "cmf.ModifyPortalContent"


class IEasyFormActionsEditorExtender(IFieldEditorExtender):
    pass


def isValidFieldName(value):
    if not ID_RE.match(value):
        raise zope.interface.Invalid(
            __(u"Please use only letters, numbers and " u"the following characters: _.")
        )
    return True


class INewAction(Schema):

    title = zope.schema.TextLine(title=__(u"Title"), required=True)

    __name__ = zope.schema.ASCIILine(
        title=__(u"Short Name"),
        description=__(u"Used for programmatic access to the field."),
        required=True,
        constraint=isValidFieldName,
    )

    factory = zope.schema.Choice(
        title=_(u"Action type"), vocabulary="EasyFormActions", required=True
    )

    @zope.interface.invariant
    def checkTitleTypes(data):
        if data.__name__ is not None and data.factory is not None:
            if (
                data.__name__ == "title"
                and data.factory.fieldcls is not zope.schema.TextLine
            ):
                raise zope.interface.Invalid(
                    __(u"The 'title' field must be a Text line (string) " u"field.")
                )


class IActionFactory(zope.schema.interfaces.IField):
    """A component that instantiates a action when called."""

    title = zope.schema.TextLine(title=__(u"Title"))


class IEasyFormActionsContext(ISchemaContext):
    """EasyForm actions view interface."""


class IActionExtender(Schema):
    fieldset(u"overrides", label=_("Overrides"), fields=["execCondition"])
    directives.read_permission(execCondition=MODIFY_PORTAL_CONTENT)
    directives.write_permission(execCondition=config.EDIT_TALES_PERMISSION)
    execCondition = zope.schema.TextLine(
        title=_(u"label_execcondition_text", default=u"Execution Condition"),
        description=_(
            u"help_execcondition_text",
            default=u"A TALES expression that will be evaluated to determine "
            u"whether or not to execute this action. Leave empty if "
            u"unneeded, and the action will be executed. Your "
            u"expression should evaluate as a boolean; return True "
            u"if you wish the action to execute. PLEASE NOTE: errors "
            u"in the evaluation of this expression will  cause an "
            u"error on form display.",
        ),
        default=u"",
        constraint=isTALES,
        required=False,
    )


class IEasyFormActionContext(IFieldContext):
    """EasyForm action content marker."""


class IActionEditForm(z3c.form.interfaces.IEditForm):
    """Marker interface for action edit forms."""


class IAction(Schema, zope.schema.interfaces.IField):
    directives.omitted(
        "description", "required", "order", "default", "missing_value", "readonly"
    )
    #     required = zope.schema.Bool(
    #         title=_('Enabled'),
    #        description=_('Tells whether a action is enabled.'),
    #         default=True)

    def onSuccess(fields, request):  # NOQA
        pass



# -*- coding: utf-8 -*-
#
# Test EasyForm initialisation and set-up
#

from collective.easyform.tests import base
from plone import api
from Products.CMFCore.utils import getToolByName

import Products


try:
    from Products.CMFPlone.utils import get_installer
except ImportError:
    # BBB for Plone 5.0 and lower.
    get_installer = None


def getAddPermission(product, name):
    """find the add permission for a meta_type"""

    name = "{0}: {1}".format(product, name)
    for mt in Products.meta_types:
        if mt["name"] == name:
            return mt["permission"]
    return ""


class TestInstallation(base.EasyFormTestCase):

    """Ensure product is properly installed"""

    def afterSetUp(self):
        base.EasyFormTestCase.afterSetUp(self)

        self.types = self.portal.portal_types
        self.properties = self.portal.portal_properties
        self.controlpanel = self.portal.portal_controlpanel

        self.metaTypes = ("EasyForm",)

    def testTypesInstalled(self):
        for t in self.metaTypes:
            self.assertTrue(t in self.types.objectIds())

    def testTypeActions(self):
        # hide properties/references tabs
        for typ in self.metaTypes:
            for act in self.types[typ].listActions():
                if act.id in ["metadata", "references"]:
                    self.assertFalse(act.visible)
                if act.id in ["actions", "fields"]:
                    self.assertTrue(act.visible)

    def testArchetypesToolCatalogRegistration(self):
        at_tool = getToolByName(self.portal, "archetype_tool", None)
        if not at_tool:
            return
        for t in self.metaTypes:
            self.assertEqual(1, len(at_tool.getCatalogsByType(t)))
            self.assertEqual("portal_catalog", at_tool.getCatalogsByType(t)[0].getId())

    def ttestControlPanelConfigletInstalled(self):
        self.assertTrue(
            "EasyForm" in [action.id for action in self.controlpanel.listActions()]
        )

    def ttestAddPermissions(self):
        """Test to make sure add permissions are as intended"""

        ADD_CONTENT_PERMISSION = "EasyForm: Add Content"
        CSA_ADD_CONTENT_PERMISSION = "EasyForm: Add Custom Scripts"
        MA_ADD_CONTENT_PERMISSION = "EasyForm: Add Mailers"
        SDA_ADD_CONTENT_PERMISSION = "EasyForm: Add Data Savers"

        self.assertEqual(
            getAddPermission("EasyForm", "EasyForm"), ADD_CONTENT_PERMISSION
        )
        self.assertEqual(
            getAddPermission("EasyForm", "Mailer Adapter"), MA_ADD_CONTENT_PERMISSION
        )
        self.assertEqual(
            getAddPermission("EasyForm", "Save Data Adapter"),
            SDA_ADD_CONTENT_PERMISSION,
        )
        self.assertEqual(
            getAddPermission("EasyForm", "Custom Script Adapter"),
            CSA_ADD_CONTENT_PERMISSION,
        )

    def testActionsInstalled(self):
        self.assertTrue(
            self.portal.portal_actions.getActionInfo("object_buttons/export")
        )
        self.assertTrue(
            self.portal.portal_actions.getActionInfo("object_buttons/import")
        )
        self.assertTrue(
            self.portal.portal_actions.getActionInfo("object_buttons/saveddata")
        )

    def test_EasyFormInDefaultPageTypes(self):
        values = api.portal.get_registry_record("plone.default_page_types")
        self.assertIn("EasyForm", values)

    def testTypeViews(self):
        self.assertEqual(
            self.types.EasyForm.getAvailableViewMethods(self.types), ("view",)
        )

    def test_upgrades(self):
        portal_setup = self.portal.portal_setup
        profile_id = "profile-collective.easyform:default"
        self.assertFalse(portal_setup.hasPendingUpgrades())
        portal_setup.setLastVersionForProfile(profile_id, "1000")
        self.assertTrue(portal_setup.hasPendingUpgrades())
        # Biggest test is: do the upgrades run without error?
        portal_setup.upgradeProfile(profile_id)
        self.assertFalse(portal_setup.hasPendingUpgrades())



# -*- coding: utf-8 -*-
#
# Integration tests specific to the api
#

from collective.easyform.actions import DummyFormView
from collective.easyform.api import filter_fields
from collective.easyform.api import filter_widgets
from collective.easyform.api import get_schema
from collective.easyform.tests import base
from plone import api


class TestFunctions(base.EasyFormTestCase):
    """Test api"""

    def afterSetUp(self):
        super(TestFunctions, self).afterSetUp()
        self.folder.invokeFactory("EasyForm", "ff1")
        self.ff1 = getattr(self.folder, "ff1")
        self.dummy_form = DummyFormView(self.ff1, self.layer["request"])
        self.dummy_form.schema = get_schema(self.ff1)
        self.dummy_form.prefix = "form"
        self.dummy_form._update()

    def test_set_fields_updates_modified(self):
        # https://github.com/collective/collective.easyform/issues/8
        # Calling set_fields should update the modification date,
        # also in the catalog.
        from collective.easyform.api import set_fields

        # Gather the original data.
        catalog = api.portal.get_tool("portal_catalog")
        path = "/".join(self.ff1.getPhysicalPath())
        orig_modified = self.ff1.modified()
        brain = catalog.unrestrictedSearchResults(path=path)[0]
        orig_counter = catalog.getCounter()
        self.assertEqual(brain.modified, orig_modified)

        # Set the fields.
        fields = get_schema(self.ff1)
        set_fields(self.ff1, fields)

        # The modification date on the form should have been updated.
        new_modified = self.ff1.modified()
        self.assertGreater(new_modified, orig_modified)

        # The catalog brain should have the new date
        brain = catalog.unrestrictedSearchResults(path=path)[0]
        self.assertEqual(brain.modified, new_modified)
        self.assertGreater(self.ff1.modified(), orig_modified)

        # The catalog counter should have been increased.
        # This helps invalidate caches because the catalogCounter ETag changes.
        self.assertEqual(catalog.getCounter(), orig_counter + 1)

    def test_set_actions_updates_modified(self):
        # https://github.com/collective/collective.easyform/issues/8
        # Calling set_actions should update the modification date,
        # also in the catalog.
        from collective.easyform.api import get_actions
        from collective.easyform.api import set_actions

        # Gather the original data.
        catalog = api.portal.get_tool("portal_catalog")
        path = "/".join(self.ff1.getPhysicalPath())
        orig_modified = self.ff1.modified()
        brain = catalog.unrestrictedSearchResults(path=path)[0]
        orig_counter = catalog.getCounter()
        self.assertEqual(brain.modified, orig_modified)

        # Set the actions.
        actions = get_actions(self.ff1)
        set_actions(self.ff1, actions)

        # The modification date on the form should have been updated.
        new_modified = self.ff1.modified()
        self.assertGreater(new_modified, orig_modified)

        # The catalog brain should have the new date
        brain = catalog.unrestrictedSearchResults(path=path)[0]
        self.assertEqual(brain.modified, new_modified)
        self.assertGreater(self.ff1.modified(), orig_modified)

        # The catalog counter should have been increased.
        # This helps invalidate caches because the catalogCounter ETag changes.
        self.assertEqual(catalog.getCounter(), orig_counter + 1)

    def test_selective_widgets(self):
        """Test selective inclusion of widgets for mail and thank you page.

        This uses filter_widgets, which needs as input:
        - a context (the form or a mailer)
        - a list of widgets
        """

        self.assertEqual(
            set(filter_widgets(self.ff1, self.dummy_form.w).keys()),
            set(["replyto", "topic", "comments"]),
        )

        self.ff1.showFields = ("topic", "comments")
        self.assertEqual(
            set(filter_widgets(self.ff1, self.dummy_form.w).keys()),
            set(["replyto", "topic", "comments"]),
        )

        self.ff1.showAll = False
        self.assertEqual(
            set(filter_widgets(self.ff1, self.dummy_form.w).keys()),
            set(["topic", "comments"]),
        )

    def test_selective_fields(self):
        """Test selective inclusion of fields for mail and thank you page.

        This uses filter_fields, which needs as input:
        - a context (the form or a mailer)
        - a list of widgets
        - (unsorted) data (user input from the request)
        - optional omit=True/False (default False)

        With omit=True, we get a list of field names to omit.
        With omit=False, we get an ordered dict of fields to include.
        """
        # Test the types of answers.
        from collections import OrderedDict as BaseDict
        from collective.easyform.api import OrderedDict

        self.assertIsInstance(
            filter_fields(self.ff1, self.dummy_form.schema, {}), OrderedDict
        )
        self.assertIsInstance(
            filter_fields(self.ff1, self.dummy_form.schema, {}), BaseDict
        )
        self.assertIsInstance(filter_fields(self.ff1, self.dummy_form.schema, {}), dict)
        self.assertIsInstance(
            filter_fields(self.ff1, self.dummy_form.schema, {}, omit=True), list
        )

        # Empty data
        self.assertEqual(
            list(filter_fields(self.ff1, self.dummy_form.schema, {}).keys()),
            [],
        )
        self.assertEqual(
            list(filter_fields(self.ff1, self.dummy_form.schema, {}, omit=True)),
            [],
        )

        # All data empty
        data = {"replyto": "", "topic": "", "comments": ""}
        self.assertEqual(
            list(filter_fields(self.ff1, self.dummy_form.schema, data).keys()),
            ["replyto", "topic", "comments"],
        )
        self.assertEqual(
            list(filter_fields(self.ff1, self.dummy_form.schema, data).values()),
            ["", "", ""],
        )
        self.assertEqual(
            list(filter_fields(self.ff1, self.dummy_form.schema, data, omit=True)),
            [],
        )
        self.ff1.includeEmpties = False
        self.assertEqual(
            list(filter_fields(self.ff1, self.dummy_form.schema, data).keys()),
            [],
        )
        self.assertEqual(
            list(filter_fields(self.ff1, self.dummy_form.schema, data, omit=True)),
            ["replyto", "topic", "comments"],
        )

        # All data filled in
        data = {"replyto": "me@example.org", "topic": "Test", "comments": "Ni"}
        self.assertEqual(
            list(filter_fields(self.ff1, self.dummy_form.schema, data).keys()),
            ["replyto", "topic", "comments"],
        )
        self.assertEqual(
            list(filter_fields(self.ff1, self.dummy_form.schema, data).values()),
            ["me@example.org", "Test", "Ni"],
        )
        self.assertEqual(
            list(filter_fields(self.ff1, self.dummy_form.schema, data, omit=True)),
            [],
        )

        # Show only specific fields, not active when showAll=True.
        self.ff1.showFields = ("no-such-field", "topic", "comments")
        self.assertEqual(
            list(filter_fields(self.ff1, self.dummy_form.schema, data).keys()),
            ["replyto", "topic", "comments"],
        )
        self.assertEqual(
            list(filter_fields(self.ff1, self.dummy_form.schema, data, omit=True)),
            [],
        )

        # Only show specific fields.
        self.ff1.showAll = False
        self.assertEqual(
            list(filter_fields(self.ff1, self.dummy_form.schema, data).keys()),
            ["topic", "comments"],
        )
        self.assertEqual(
            list(filter_fields(self.ff1, self.dummy_form.schema, data, omit=True)),
            ["replyto"],
        )

        # The order of showFields is kept.
        # Note: currently the order is used on the mail template,
        # but not on the thanks page.
        self.ff1.showFields = ("comments", "no-such-field", "topic")
        self.assertEqual(
            list(filter_fields(self.ff1, self.dummy_form.schema, data).keys()),
            ["comments", "topic"],
        )
        self.assertEqual(
            list(filter_fields(self.ff1, self.dummy_form.schema, data, omit=True)),
            ["replyto"],
        )
